import json
import pathlib
import re
from unittest import mock
from unittest.mock import ANY as ANY_TEST_ARG

import pytest
from pytest_mock import MockerFixture

import great_expectations.expectations as gxe
from great_expectations.analytics.events import (
    DomainObjectAllDeserializationEvent,
    ValidationDefinitionCreatedEvent,
    ValidationDefinitionDeletedEvent,
)
from great_expectations.compatibility.pydantic import ValidationError
from great_expectations.core.batch_definition import BatchDefinition
from great_expectations.core.expectation_suite import ExpectationSuite
from great_expectations.core.factory.validation_definition_factory import (
    ValidationDefinitionFactory,
)
from great_expectations.core.validation_definition import ValidationDefinition
from great_expectations.data_context.data_context.abstract_data_context import (
    AbstractDataContext,
)
from great_expectations.data_context.data_context.cloud_data_context import (
    CloudDataContext,
)
from great_expectations.data_context.data_context.file_data_context import (
    FileDataContext,
)
from great_expectations.data_context.store.validation_definition_store import (
    ValidationDefinitionStore,
)
from great_expectations.exceptions import DataContextError
from tests.conftest import random_name


@pytest.fixture
def validation_definition(mocker: MockerFixture) -> ValidationDefinition:
    batch_definition = mocker.Mock(spec=BatchDefinition)
    suite = mocker.Mock(spec=ExpectationSuite)
    return ValidationDefinition(
        name="test-validation",
        data=batch_definition,
        suite=suite,
    )


@pytest.fixture
def validation_definition_json(validation_definition: ValidationDefinition) -> str:
    return json.dumps(
        {
            "name": validation_definition.name,
            "id": None,
            "data": {
                "datasource": {
                    "name": "test-datasource",
                    "id": "f4b3d8f2-7e4d-4f0b-9b90-6f711d0e3e2f",
                },
                "asset": {
                    "name": "test-asset",
                    "id": "10b7e57d-958b-4d28-aa1d-e89bc0401eea",
                },
                "batch_definition": {
                    "name": "test-batch-config",
                    "id": "bcd13e3e-3e3e-4e3e-8e3e-3e3e3e3e3e3e",
                },
            },
            "suite": {
                "name": "test-expectation-suite",
                "id": "06022df2-6699-49b2-8c03-df77678363a0",
            },
        }
    )


@pytest.mark.unit
def test_validation_definition_factory_get_uses_store_get(
    mocker: MockerFixture, validation_definition: ValidationDefinition
):
    # Arrange
    name = validation_definition.name
    store = mocker.Mock(spec=ValidationDefinitionStore)
    store.has_key.return_value = True
    key = store.get_key.return_value
    store.get.return_value = validation_definition
    factory = ValidationDefinitionFactory(store=store)

    # Act
    result = factory.get(name=name)

    # Assert
    store.get.assert_called_once_with(key=key)
    assert result == validation_definition


@pytest.mark.unit
def test_validation_definition_factory_get_raises_error_on_missing_key(
    mocker: MockerFixture,
    validation_definition: ValidationDefinition,
):
    # Arrange
    name = validation_definition.name
    store = mocker.Mock(spec=ValidationDefinitionStore)
    store.has_key.return_value = False
    store.get.return_value = validation_definition
    factory = ValidationDefinitionFactory(store=store)

    # Act
    with pytest.raises(
        DataContextError, match=f"ValidationDefinition with name {name} was not found."
    ):
        factory.get(name=name)

    # Assert
    store.get.assert_not_called()


@pytest.mark.unit
def test_validation_definition_factory_add_uses_store_add(
    mocker: MockerFixture, validation_definition: ValidationDefinition
):
    # Arrange
    store = mocker.Mock(spec=ValidationDefinitionStore)
    store.has_key.return_value = False
    key = store.get_key.return_value
    factory = ValidationDefinitionFactory(store=store)
    store.get.return_value = validation_definition

    # Act
    factory.add(validation=validation_definition)

    # Assert
    store.add.assert_called_once_with(key=key, value=validation_definition)


@pytest.mark.unit
def test_validation_definition_factory_add_raises_for_duplicate_key(
    mocker: MockerFixture,
    validation_definition: ValidationDefinition,
):
    # Arrange
    name = validation_definition.name
    store = mocker.Mock(spec=ValidationDefinitionStore)
    store.has_key.return_value = True
    factory = ValidationDefinitionFactory(store=store)

    # Act
    with pytest.raises(
        DataContextError,
        match=f"Cannot add ValidationDefinition with name {name} because it already exists.",
    ):
        factory.add(validation=validation_definition)

    # Assert
    store.add.assert_not_called()


@pytest.mark.unit
def test_validation_definition_factory_delete_uses_store_remove_key(
    mocker: MockerFixture,
    validation_definition: ValidationDefinition,
):
    # Arrange
    store = mocker.Mock(spec=ValidationDefinitionStore)
    store.has_key.return_value = True
    key = store.get_key.return_value
    factory = ValidationDefinitionFactory(store=store)

    # Act
    factory.delete(name="my_validation_def")

    # Assert
    store.remove_key.assert_called_once_with(
        key=key,
    )


@pytest.mark.unit
def test_validation_definition_factory_delete_raises_for_missing_validation(
    mocker: MockerFixture,
    validation_definition: ValidationDefinition,
):
    # Arrange
    name = validation_definition.name
    store = mocker.Mock(spec=ValidationDefinitionStore)
    store.has_key.return_value = False
    factory = ValidationDefinitionFactory(store=store)

    # Act
    with pytest.raises(
        DataContextError,
        match=f"Cannot delete ValidationDefinition with name {name} because it cannot be found.",
    ):
        factory.delete(name=name)

    # Assert
    store.remove_key.assert_not_called()


@pytest.mark.filesystem
def test_validation_definition_factory_is_initialized_with_context_filesystem(
    empty_data_context: FileDataContext,
):
    assert isinstance(empty_data_context.validation_definitions, ValidationDefinitionFactory)


@pytest.mark.cloud
def test_validation_definition_factory_is_initialized_with_context_cloud(
    empty_cloud_data_context: CloudDataContext,
):
    assert isinstance(empty_cloud_data_context.validation_definitions, ValidationDefinitionFactory)


@pytest.mark.filesystem
def test_validation_definition_factory_add_success_filesystem(
    empty_data_context: FileDataContext,
    validation_definition: ValidationDefinition,
    validation_definition_json: str,
    mocker: MockerFixture,
):
    _test_validation_definition_factory_add_success(
        mocker=mocker,
        context=empty_data_context,
        validation_definition=validation_definition,
        validation_definition_json=validation_definition_json,
    )


@pytest.mark.cloud
def test_validation_definition_factory_add_success_cloud(
    empty_cloud_context_fluent: CloudDataContext,
    validation_definition: ValidationDefinition,
    validation_definition_json: str,
    mocker: MockerFixture,
):
    _test_validation_definition_factory_add_success(
        mocker=mocker,
        context=empty_cloud_context_fluent,
        validation_definition=validation_definition,
        validation_definition_json=validation_definition_json,
    )


def _test_validation_definition_factory_add_success(
    mocker: MockerFixture,
    context: AbstractDataContext,
    validation_definition: ValidationDefinition,
    validation_definition_json: str,
):
    # Arrange
    name = validation_definition.name
    with pytest.raises(
        DataContextError, match=f"ValidationDefinition with name {name} was not found."
    ):
        context.validation_definitions.get(name)

    # Act
    with mocker.patch.object(ValidationDefinition, "json", return_value=validation_definition_json):
        created_validation = context.validation_definitions.add(validation=validation_definition)

    # Assert
    validation_names = {
        key.to_tuple()[0] for key in context.validation_definition_store.list_keys()
    }
    assert created_validation.name in validation_names


@pytest.mark.filesystem
def test_validation_definition_factory_delete_success_filesystem(
    empty_data_context: FileDataContext,
    validation_definition: ValidationDefinition,
    validation_definition_json: str,
    mocker: MockerFixture,
):
    _test_validation_definition_factory_delete_success(
        mocker=mocker,
        context=empty_data_context,
        validation_definition=validation_definition,
        validation_definition_json=validation_definition_json,
    )


@pytest.mark.cloud
def test_validation_definition_factory_delete_success_cloud(
    empty_cloud_context_fluent: CloudDataContext,
    validation_definition: ValidationDefinition,
    validation_definition_json: str,
    mocker: MockerFixture,
):
    _test_validation_definition_factory_delete_success(
        mocker=mocker,
        context=empty_cloud_context_fluent,
        validation_definition=validation_definition,
        validation_definition_json=validation_definition_json,
    )


def _test_validation_definition_factory_delete_success(
    mocker: MockerFixture,
    context: AbstractDataContext,
    validation_definition: ValidationDefinition,
    validation_definition_json: str,
):
    # Arrange
    name = validation_definition.name

    with mocker.patch.object(ValidationDefinition, "json", return_value=validation_definition_json):
        validation_definition = context.validation_definitions.add(validation=validation_definition)

    # Act
    with mocker.patch.object(ValidationDefinition, "parse_raw", return_value=validation_definition):
        context.validation_definitions.delete(name=name)

    # Assert
    with pytest.raises(
        DataContextError,
        match=f"ValidationDefinition with name {name} was not found.",
    ):
        context.validation_definitions.get(name)


@pytest.mark.parametrize(
    "context_fixture_name",
    [
        pytest.param("empty_cloud_context_fluent", id="cloud", marks=pytest.mark.cloud),
        pytest.param("in_memory_runtime_context", id="ephemeral", marks=pytest.mark.big),
        pytest.param("empty_data_context", id="filesystem", marks=pytest.mark.filesystem),
    ],
)
def test_validation_definition_factory_all(
    context_fixture_name: str, request: pytest.FixtureRequest
):
    context: AbstractDataContext = request.getfixturevalue(context_fixture_name)

    # Arrange
    ds = context.data_sources.add_pandas("my_datasource")
    asset = ds.add_csv_asset("my_asset", "data.csv")  # type: ignore[arg-type] # FIXME CoP
    suite = context.suites.add(ExpectationSuite(name="my_suite"))
    validation_definition_a = ValidationDefinition(
        name="validation definition a",
        data=asset.add_batch_definition("a"),
        suite=suite,
    )
    validation_definition_b = ValidationDefinition(
        name="validation definition b",
        data=asset.add_batch_definition("b"),
        suite=suite,
    )

    context.validation_definitions.add(validation=validation_definition_a)
    context.validation_definitions.add(validation=validation_definition_b)

    # Act
    result = context.validation_definitions.all()
    result = sorted(result, key=lambda x: x.name)

    # Assert
    assert [r.name for r in result] == [validation_definition_a.name, validation_definition_b.name]
    assert result == [validation_definition_a, validation_definition_b]


@pytest.mark.unit
def test_validation_definition_factory_all_with_bad_config(
    in_memory_runtime_context: AbstractDataContext,
    mocker: MockerFixture,
):
    mocker.patch(
        "great_expectations.core.validation_definition.ValidationDefinition.Config.validate_assignment",
        False,
    )
    analytics_submit_mock = mocker.patch(
        "great_expectations.data_context.store.store.submit_analytics_event"
    )
    context: AbstractDataContext = in_memory_runtime_context

    # Arrange
    ds = context.data_sources.add_pandas("my_datasource")
    asset = ds.add_csv_asset("my_asset", "data.csv")  # type: ignore[arg-type] # FIXME CoP
    suite = context.suites.add(ExpectationSuite(name="my_suite"))

    validation_definition_1 = ValidationDefinition(
        name="validation1",
        data=asset.add_batch_definition("1"),
        suite=suite,
    )
    context.validation_definitions.add(validation=validation_definition_1)

    validation_definition_2 = ValidationDefinition(
        name="validation2",
        data=asset.add_batch_definition("2"),
        suite=suite,
    )
    context.validation_definitions.add(validation=validation_definition_2)

    # Verify our validation definitions are added
    assert sorted(context.validation_definitions.all(), key=lambda cp: cp.name) == [
        validation_definition_1,
        validation_definition_2,
    ]

    # Make validation_definition_2 invalid. Pydantic will validate the object at creation time
    # but we can invalidate via assignment because of the monkeypatch at the top of this test.
    validation_definition_2.suite = None  # type: ignore[assignment] # done intentionally for test
    validation_definition_2.save()

    # Act
    result = context.validation_definitions.all()

    # Assert
    assert result == [validation_definition_1]
    analytics_submit_mock.assert_called_once_with(
        DomainObjectAllDeserializationEvent(
            error_type=ANY_TEST_ARG,
            store_name="ValidationDefinitionStore",
        )
    )
    analytics_submit_args = analytics_submit_mock.call_args[0][0]
    assert re.match("pydantic.*ValidationError", analytics_submit_args.error_type)


@pytest.mark.filesystem
def test_validation_definition_factory_round_trip(
    empty_data_context: FileDataContext,
    validation_definition: ValidationDefinition,
):
    # Arrange
    context = empty_data_context

    ds = context.data_sources.add_pandas("my_ds")
    csv_path = pathlib.Path(
        __file__, "..", "..", "..", "test_sets", "quickstart", "yellow_tripdata_sample_2022-01.csv"
    ).resolve()
    assert csv_path.exists()
    asset = ds.add_csv_asset("my_asset", filepath_or_buffer=csv_path)

    batch_definition = asset.add_batch_definition("my_batch_def")
    suite = context.suites.add(
        ExpectationSuite(
            name="my_suite",
            expectations=[
                gxe.ExpectColumnValuesToBeBetween(
                    column="passenger_count", min_value=0, max_value=10
                ),
                gxe.ExpectColumnMeanToBeBetween(
                    column="passenger_count",
                    min_value=0,
                    max_value=1,
                ),
            ],
        )
    )

    # Act
    validation_definition = ValidationDefinition(
        name="my_validation_def", data=batch_definition, suite=suite
    )
    persisted_validation_definition = context.validation_definitions.add(
        validation=validation_definition
    )
    retrieved_validation_definition = context.validation_definitions.get(
        name=validation_definition.name
    )

    # Assert
    # Suite equality is a bit finnicky, so we just check the JSON representation
    assert persisted_validation_definition.json() == retrieved_validation_definition.json()


class TestValidationDefinitionFactoryAddOrUpdate:
    def _build_batch_definition(self, context: AbstractDataContext):
        name = random_name()
        ds = context.data_sources.add_pandas(name=name)
        asset = ds.add_csv_asset(name=name, filepath_or_buffer=pathlib.Path("data.csv"))
        return asset.add_batch_definition(name=name)

    def _build_suite(self) -> ExpectationSuite:
        return ExpectationSuite(
            name=random_name(),
            expectations=[
                gxe.ExpectColumnValuesToBeBetween(
                    column="passenger_count", min_value=0, max_value=10
                )
            ],
        )

    def test_add_new_validation(self, data_context: AbstractDataContext) -> None:
        # arrange
        vd_name = random_name()
        batch_def = self._build_batch_definition(data_context)
        suite = self._build_suite()
        vd = ValidationDefinition(
            name=vd_name,
            data=batch_def,
            suite=data_context.suites.add(suite),
        )

        # act
        created_vd = data_context.validation_definitions.add_or_update(validation=vd)

        # assert
        assert created_vd.id
        data_context.validation_definitions.get(vd_name)

    def test_add_new_validation_with_new_suite(self, data_context: AbstractDataContext) -> None:
        # arrange
        vd_name = random_name()
        batch_def = self._build_batch_definition(data_context)
        suite = self._build_suite()
        vd = ValidationDefinition(
            name=vd_name,
            data=batch_def,
            suite=suite,
        )

        # act
        created_vd = data_context.validation_definitions.add_or_update(validation=vd)

        # assert
        assert created_vd.id
        data_context.validation_definitions.get(vd_name)

    def test_update_expectation_suite(self, data_context: AbstractDataContext) -> None:
        # arrange
        vd_name = random_name()
        batch_def = self._build_batch_definition(data_context)
        suite = self._build_suite()
        vd = ValidationDefinition(
            name=vd_name,
            data=batch_def,
            suite=data_context.suites.add(suite),
        )
        existing_vd = data_context.validation_definitions.add(validation=vd)

        # act
        vd.suite.expectations = [
            gxe.ExpectColumnMaxToBeBetween(column="passenger_count", min_value=0, max_value=5)
        ]

        updated_vd = data_context.validation_definitions.add_or_update(validation=vd)

        # assert
        assert updated_vd.id == existing_vd.id
        assert len(updated_vd.suite.expectations) == 1 and isinstance(
            updated_vd.suite.expectations[0], gxe.ExpectColumnMaxToBeBetween
        )
        data_context.validation_definitions.get(vd_name)

    def test_replace_expectation_suite(self, data_context: AbstractDataContext) -> None:
        # arrange
        vd_name = random_name()
        batch_def = self._build_batch_definition(data_context)
        suite = data_context.suites.add(self._build_suite())
        vd = ValidationDefinition(
            name=vd_name,
            data=batch_def,
            suite=suite,
        )
        existing_vd = data_context.validation_definitions.add(validation=vd)

        # act
        new_suite = data_context.suites.add(self._build_suite())
        new_vd = ValidationDefinition(
            name=vd_name,
            data=batch_def,
            suite=new_suite,
        )
        updated_vd = data_context.validation_definitions.add_or_update(validation=new_vd)

        # assert
        assert updated_vd.suite.id != existing_vd.suite.id  # New suite should have a different ID
        assert updated_vd.data.id == existing_vd.data.id
        assert updated_vd.id == existing_vd.id
        data_context.validation_definitions.get(vd_name)

    @pytest.mark.xfail(
        raises=ValidationError,
        reason="GX-481: Changing a BatchDefinition breaks loading a Validation Definition.",
        strict=True,
    )
    def test_update_batch_definition(self, data_context: AbstractDataContext) -> None:
        # arrange
        vd_name = random_name()
        batch_def = self._build_batch_definition(data_context)
        suite = data_context.suites.add(self._build_suite())
        vd = ValidationDefinition(
            name=vd_name,
            data=batch_def,
            suite=suite,
        )
        data_context.validation_definitions.add(validation=vd)

        # act
        new_batch_def_name = random_name()
        batch_def.name = new_batch_def_name
        new_vd = ValidationDefinition(
            name=vd_name,
            data=batch_def,
            suite=suite,
        )
        updated_vd = data_context.validation_definitions.add_or_update(validation=new_vd)

        # assert
        assert updated_vd.data.name == new_batch_def_name
        data_context.validation_definitions.get(vd_name)

    def test_replace_batch_definition(self, data_context: AbstractDataContext) -> None:
        # arrange
        vd_name = random_name()
        batch_def = self._build_batch_definition(data_context)
        suite = data_context.suites.add(self._build_suite())
        vd = ValidationDefinition(
            name=vd_name,
            data=batch_def,
            suite=suite,
        )
        existing_vd = data_context.validation_definitions.add(validation=vd)

        # act
        new_batch_def = self._build_batch_definition(data_context)
        new_vd = ValidationDefinition(
            name=vd_name,
            data=new_batch_def,
            suite=suite,
        )
        updated_vd = data_context.validation_definitions.add_or_update(validation=new_vd)

        # assert
        assert updated_vd.suite.id == existing_vd.suite.id
        assert (
            updated_vd.data.id != existing_vd.data.id
        )  # New batch definition should have a different ID
        assert updated_vd.data.id  # should not be None
        assert updated_vd.data.id == new_batch_def.id
        assert updated_vd.id == existing_vd.id
        data_context.validation_definitions.get(vd_name)

    def test_add_or_update_is_idempotent(self, data_context: AbstractDataContext) -> None:
        # arrange
        vd_name = random_name()
        batch_def = self._build_batch_definition(data_context)
        suite = self._build_suite()
        vd = ValidationDefinition(
            name=vd_name,
            data=batch_def,
            suite=suite,
        )

        # act
        vd_1 = data_context.validation_definitions.add_or_update(validation=vd)
        vd_2 = data_context.validation_definitions.add_or_update(validation=vd)
        vd_3 = data_context.validation_definitions.add_or_update(validation=vd)

        # assert
        assert vd_1 == vd_2 == vd_3


class TestValidationDefinitionFactoryAnalytics:
    def test_validation_definition_factory_add_emits_event(
        self, data_context: AbstractDataContext
    ) -> None:
        # Arrange
        ds = data_context.data_sources.add_pandas("my_datasource")
        asset = ds.add_csv_asset("my_asset", pathlib.Path("data.csv"))
        batch_def = asset.add_batch_definition("my_batch_definition")
        suite = data_context.suites.add(ExpectationSuite(name="my_suite"))

        validation_definition = ValidationDefinition(
            name="validation_def", data=batch_def, suite=suite
        )

        # Act
        with mock.patch(
            "great_expectations.core.factory.validation_definition_factory.submit_event",
            autospec=True,
        ) as mock_submit:
            _ = data_context.validation_definitions.add(validation=validation_definition)

        # Assert
        mock_submit.assert_called_once_with(
            event=ValidationDefinitionCreatedEvent(
                validation_definition_id=validation_definition.id
            )
        )

    def test_validation_definition_factory_delete_emits_event(
        self, data_context: AbstractDataContext
    ) -> None:
        # Arrange
        ds = data_context.data_sources.add_pandas("my_datasource")
        asset = ds.add_csv_asset("my_asset", pathlib.Path("data.csv"))
        batch_def = asset.add_batch_definition("my_batch_definition")
        suite = data_context.suites.add(ExpectationSuite(name="my_suite"))

        name = "validation_def"
        validation_definition = data_context.validation_definitions.add(
            validation=ValidationDefinition(name=name, data=batch_def, suite=suite)
        )

        # Act
        with mock.patch(
            "great_expectations.core.factory.validation_definition_factory.submit_event",
            autospec=True,
        ) as mock_submit:
            data_context.validation_definitions.delete(name=name)

        # Assert
        mock_submit.assert_called_once_with(
            event=ValidationDefinitionDeletedEvent(
                validation_definition_id=validation_definition.id
            )
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from great_expectations.expectations.expectation import (
    ColumnMapExpectation,
    render_suite_parameter_string,
)
from great_expectations.render import LegacyRendererType, RenderedStringTemplateContent
from great_expectations.render.renderer.renderer import renderer
from great_expectations.render.renderer_configuration import (
    RendererConfiguration,
    RendererValueType,
)
from great_expectations.render.util import (
    num_to_str,
    parse_row_condition_string_pandas_engine,
    substitute_none_for_missing,
)

if TYPE_CHECKING:
    from great_expectations.core import (
        ExpectationValidationResult,
    )
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )
    from great_expectations.render.renderer_configuration import AddParamArgs

try:
    import sqlalchemy as sa  # noqa: F401, TID251 # FIXME CoP
except ImportError:
    pass


class ExpectColumnValuesToBeJsonParseable(ColumnMapExpectation):
    """Expect the column entries to be data written in JavaScript Object Notation.

    ExpectColumnValuesToBeJsonParseable is a \
    Column Map Expectation.

    Args:
        column (str): \
            The column name.

    Keyword Args:
        mostly (None or a float between 0 and 1): \
            Successful if at least mostly fraction of values match the expectation. \
            For more detail, see [mostly](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#mostly).

    Other Parameters:
        result_format (str or None): \
            Which output mode to use: BOOLEAN_ONLY, BASIC, COMPLETE, or SUMMARY. \
            For more detail, see [result_format](https://docs.greatexpectations.io/docs/reference/expectations/result_format).
        catch_exceptions (boolean or None): \
            If True, then catch exceptions and include them as part of the result object. \
            For more detail, see [catch_exceptions](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#catch_exceptions).
        meta (dict or None): \
            A JSON-serializable dictionary (nesting allowed) that will be included in the output without \
            modification. For more detail, see [meta](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#meta).

    Returns:
        An [ExpectationSuiteValidationResult](https://docs.greatexpectations.io/docs/terms/validation_result)

        Exact fields vary depending on the values passed to result_format, catch_exceptions, and meta.

    See Also:
        [ExpectColumnValuesToMatchJsonSchema](https://greatexpectations.io/expectations/expect_column_values_to_match_json_schema)
    """  # noqa: E501 # FIXME CoP

    # This dictionary contains metadata for display in the public gallery
    library_metadata = {
        "maturity": "production",
        "tags": ["core expectation", "column map expectation"],
        "contributors": ["@great_expectations"],
        "requirements": [],
        "has_full_test_suite": True,
        "manually_reviewed_code": True,
    }

    map_metric = "column_values.json_parseable"
    success_keys = ("mostly",)
    args_keys = ("column",)

    @classmethod
    def _prescriptive_template(
        cls,
        renderer_configuration: RendererConfiguration,
    ) -> RendererConfiguration:
        add_param_args: AddParamArgs = (
            ("column", RendererValueType.STRING),
            ("mostly", RendererValueType.NUMBER),
        )
        for name, param_type in add_param_args:
            renderer_configuration.add_param(name=name, param_type=param_type)

        params = renderer_configuration.params

        template_str = "values must be parseable as JSON"

        if params.mostly and params.mostly.value < 1.0:
            renderer_configuration = cls._add_mostly_pct_param(
                renderer_configuration=renderer_configuration
            )
            template_str += ", at least $mostly_pct % of the time."
        else:
            template_str += "."

        if renderer_configuration.include_column_name:
            template_str = f"$column {template_str}"

        renderer_configuration.template_str = template_str

        return renderer_configuration

    @classmethod
    @renderer(renderer_type=LegacyRendererType.PRESCRIPTIVE)
    @render_suite_parameter_string
    def _prescriptive_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ):
        runtime_configuration = runtime_configuration or {}
        include_column_name = runtime_configuration.get("include_column_name") is not False
        styling = runtime_configuration.get("styling")
        params = substitute_none_for_missing(
            configuration.kwargs,
            ["column", "mostly", "row_condition", "condition_parser"],
        )

        template_str = "values must be parseable as JSON"

        if params["mostly"] is not None:
            if isinstance(params["mostly"], (int, float)) and params["mostly"] < 1.0:
                params["mostly_pct"] = num_to_str(params["mostly"] * 100, no_scientific=True)
                # params["mostly_pct"] = "{:.14f}".format(params["mostly"]*100).rstrip("0").rstrip(".")  # noqa: E501 # FIXME CoP
            template_str += ", at least $mostly_pct % of the time."
        else:
            template_str += "."

        if include_column_name:
            template_str = f"$column {template_str}"

        if params["row_condition"] is not None:
            (
                conditional_template_str,
                conditional_params,
            ) = parse_row_condition_string_pandas_engine(params["row_condition"])
            template_str = f"{conditional_template_str}, then {template_str}"
            params.update(conditional_params)

        return [
            RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": template_str,
                        "params": params,
                        "styling": styling,
                    },
                }
            )
        ]



from __future__ import annotations

import logging
import warnings
from pprint import pformat as pf
from typing import TYPE_CHECKING, Any, Generator
from unittest import mock

import pytest
from pytest import param

from great_expectations.compatibility import sqlalchemy
from great_expectations.compatibility.sqlalchemy import sqlalchemy as sa
from great_expectations.datasource.fluent import GxDatasourceWarning, SQLDatasource
from great_expectations.datasource.fluent.sql_datasource import (
    DEFAULT_QUOTE_CHARACTERS,
    TableAsset,
    to_lower_if_not_quoted,
)
from great_expectations.execution_engine import SqlAlchemyExecutionEngine

if TYPE_CHECKING:
    from pytest_mock import MockerFixture

    from great_expectations.data_context import (
        EphemeralDataContext,
    )


LOGGER = logging.getLogger(__name__)


@pytest.fixture
def create_engine_spy(mocker: MockerFixture) -> Generator[mock.MagicMock, None, None]:  # noqa: TID251 # FIXME CoP
    spy = mocker.spy(sa, "create_engine")
    yield spy
    if not spy.call_count:
        LOGGER.warning("SQLAlchemy create_engine was not called")


@pytest.fixture
def gx_sqlalchemy_execution_engine_spy(
    mocker: MockerFixture, monkeypatch: pytest.MonkeyPatch
) -> Generator[mock.MagicMock, None, None]:  # noqa: TID251 # FIXME CoP
    """
    Mock the SQLDatasource.execution_engine_type property to return a spy so that what would be passed to
    the GX SqlAlchemyExecutionEngine constructor can be inspected.

    NOTE: This is not exactly what gets passed to the sqlalchemy.engine.create_engine() function, but it is close.
    """  # noqa: E501 # FIXME CoP
    spy = mocker.Mock(spec=SqlAlchemyExecutionEngine)
    monkeypatch.setattr(SQLDatasource, "execution_engine_type", spy)
    yield spy
    if not spy.call_count:
        LOGGER.warning("SqlAlchemyExecutionEngine.__init__() was not called")


@pytest.fixture
def create_engine_fake(monkeypatch: pytest.MonkeyPatch) -> None:
    """Monkeypatch sqlalchemy.create_engine to always return a in-memory sqlite engine."""
    in_memory_sqlite_engine = sa.create_engine("sqlite:///")

    def _fake_create_engine(*args, **kwargs) -> sa.engine.Engine:
        LOGGER.info(f"Mock create_engine called with {args=} {kwargs=}")
        return in_memory_sqlite_engine

    monkeypatch.setattr(sa, "create_engine", _fake_create_engine, raising=True)


@pytest.mark.unit
@pytest.mark.parametrize(
    "ds_kwargs",
    [
        param(
            dict(
                connection_string="sqlite:///",
            ),
            id="connection_string only",
        ),
        param(
            dict(
                connection_string="sqlite:///",
                kwargs={"isolation_level": "SERIALIZABLE"},
            ),
            id="no subs + kwargs",
        ),
        param(
            dict(
                connection_string="${MY_CONN_STR}",
                kwargs={"isolation_level": "SERIALIZABLE"},
            ),
            id="subs + kwargs",
        ),
        param(
            dict(
                connection_string="sqlite:///",
                create_temp_table=True,
            ),
            id="create_temp_table=True",
        ),
        param(
            dict(
                connection_string="sqlite:///",
                create_temp_table=False,
            ),
            id="create_temp_table=False",
        ),
    ],
)
class TestConfigPasstrough:
    def test_kwargs_passed_to_create_engine(
        self,
        create_engine_spy: mock.MagicMock,  # noqa: TID251 # FIXME CoP
        monkeypatch: pytest.MonkeyPatch,
        ephemeral_context_with_defaults: EphemeralDataContext,
        ds_kwargs: dict,
        filter_gx_datasource_warnings: None,
    ):
        monkeypatch.setenv("MY_CONN_STR", "sqlite:///")

        context = ephemeral_context_with_defaults
        ds = context.data_sources.add_or_update_sql(name="my_datasource", **ds_kwargs)
        print(ds)
        ds.test_connection()

        create_engine_spy.assert_called_once_with(
            "sqlite:///",
            **{
                **ds.dict(include={"kwargs"}, exclude_unset=False)["kwargs"],
                **ds_kwargs.get("kwargs", {}),
            },
        )

    def test_ds_config_passed_to_gx_sqlalchemy_execution_engine(
        self,
        gx_sqlalchemy_execution_engine_spy: mock.MagicMock,  # noqa: TID251 # FIXME CoP
        monkeypatch: pytest.MonkeyPatch,
        ephemeral_context_with_defaults: EphemeralDataContext,
        ds_kwargs: dict,
        filter_gx_datasource_warnings: None,
    ):
        monkeypatch.setenv("MY_CONN_STR", "sqlite:///")

        context = ephemeral_context_with_defaults
        ds = context.data_sources.add_or_update_sql(name="my_datasource", **ds_kwargs)
        print(ds)
        gx_execution_engine: SqlAlchemyExecutionEngine = ds.get_execution_engine()
        print(f"{gx_execution_engine=}")

        expected_args: dict[str, Any] = {
            # kwargs that we expect are passed to SqlAlchemyExecutionEngine
            # including datasource field default values
            **ds.dict(
                exclude_unset=False,
                exclude={"kwargs", *ds_kwargs.keys(), *ds._get_exec_engine_excludes()},
            ),
            **{k: v for k, v in ds_kwargs.items() if k not in ["kwargs"]},
            **ds_kwargs.get("kwargs", {}),
            # config substitution should have been performed
            **ds.dict(include={"connection_string"}, config_provider=ds._config_provider),
        }
        assert "create_temp_table" in expected_args

        print(f"\nExpected SqlAlchemyExecutionEngine arguments:\n{pf(expected_args)}")
        gx_sqlalchemy_execution_engine_spy.assert_called_once_with(**expected_args)


@pytest.mark.unit
def test_table_quoted_name_type_does_not_exist(
    mocker,
):
    """
    DBMS entity names (table, column, etc.) must adhere to correct case insensitivity standards.  All upper case is
    standard for Oracle, DB2, and Snowflake, while all lowercase is standard for SQLAlchemy; hence, proper conversion to
    quoted names must occur.  This test ensures that mechanism for detection of non-existent table_nam" works correctly.
    """  # noqa: E501 # FIXME CoP
    table_names_in_dbms_schema: list[str] = [
        "table_name_0",
        "table_name_1",
        "table_name_2",
        "table_name_3",
    ]

    with mock.patch(
        "great_expectations.datasource.fluent.sql_datasource.TableAsset.datasource",
        new_callable=mock.PropertyMock,
        return_value=SQLDatasource(
            name="my_snowflake_datasource",
            connection_string="snowflake://<user_login_name>:<password>@<account_identifier>/<database_name>/<schema_name>?warehouse=<warehouse_name>&role=<role_name>",
        ),
    ):
        table_asset = TableAsset(
            name="my_table_asset",
            table_name="nonexistent_table_name",
            schema_name="my_schema",
        )
        assert table_asset.table_name not in table_names_in_dbms_schema


@pytest.mark.unit
def test_table_quoted_name_type_all_upper_case_normalizion_is_noop():
    """
    DBMS entity names (table, column, etc.) must adhere to correct case insensitivity standards.  All upper case is
    standard for Oracle, DB2, and Snowflake, while all lowercase is standard for SQLAlchemy; hence, proper conversion to
    quoted names must occur.  This test ensures that all upper case entity usage does not undergo any conversion.
    """  # noqa: E501 # FIXME CoP
    table_names_in_dbms_schema: list[str] = [
        "ACTORS",
        "ARTISTS",
        "ATHLETES",
        "BUSINESS_PEOPLE",
        "HEALTHCARE_WORKERS",
        "ENGINEERS",
        "LAWYERS",
        "MUSICIANS",
        "SCIENTISTS",
        "LITERARY_PROFESSIONALS",
    ]

    asset_name: str
    table_name: str

    with mock.patch(
        "great_expectations.datasource.fluent.sql_datasource.TableAsset.datasource",
        new_callable=mock.PropertyMock,
        return_value=SQLDatasource(
            name="my_snowflake_datasource",
            connection_string="snowflake://<user_login_name>:<password>@<account_identifier>/<database_name>/<schema_name>?warehouse=<warehouse_name>&role=<role_name>",
        ),
    ):
        for table_name in table_names_in_dbms_schema:
            asset_name = f"{table_name}_asset"
            table_asset = TableAsset(
                name=asset_name,
                table_name=table_name,
                schema_name="my_schema",
            )
            assert str(table_asset.table_name) == table_name
            assert str(table_asset.table_name.casefold()) != table_name
            assert isinstance(table_asset.table_name, sqlalchemy.quoted_name)
            assert table_asset.table_name in table_names_in_dbms_schema


@pytest.mark.unit
def test_table_quoted_name_type_all_lower_case_normalizion_full():
    """
    DBMS entity names (table, column, etc.) must adhere to correct case insensitivity standards.  All upper case is
    standard for Oracle, DB2, and Snowflake, while all lowercase is standard for SQLAlchemy; hence, proper conversion to
    quoted names must occur.  This test ensures that all lower case entity usage undergo conversion to quoted literals.
    """  # noqa: E501 # FIXME CoP
    table_names_in_dbms_schema: list[str] = [
        "actors",
        "artists",
        "athletes",
        "business_people",
        "healthcare_workers",
        "engineers",
        "lawyers",
        "musicians",
        "scientists",
        "literary_professionals",
    ]

    name: str

    quoted_table_names: list[sqlalchemy.quoted_name] = [
        sqlalchemy.quoted_name(value="actors", quote=True),
        sqlalchemy.quoted_name(value="artists", quote=True),
        sqlalchemy.quoted_name(value="athletes", quote=True),
        sqlalchemy.quoted_name(value="business_people", quote=True),
        sqlalchemy.quoted_name(value="healthcare_workers", quote=True),
        sqlalchemy.quoted_name(value="engineers", quote=True),
        sqlalchemy.quoted_name(value="lawyers", quote=True),
        sqlalchemy.quoted_name(value="musicians", quote=True),
        sqlalchemy.quoted_name(value="scientists", quote=True),
        sqlalchemy.quoted_name(value="literary_professionals", quote=True),
    ]

    asset_name: str
    table_name: str

    with mock.patch(
        "great_expectations.datasource.fluent.sql_datasource.TableAsset.datasource",
        new_callable=mock.PropertyMock,
        return_value=SQLDatasource(
            name="my_snowflake_datasource",
            connection_string="snowflake://<user_login_name>:<password>@<account_identifier>/<database_name>/<schema_name>?warehouse=<warehouse_name>&role=<role_name>",
        ),
    ):
        for table_name in table_names_in_dbms_schema:
            asset_name = f"{table_name}_asset"
            table_asset = TableAsset(
                name=asset_name,
                table_name=table_name,
                schema_name="my_schema",
            )
            assert str(table_asset.table_name) == table_name
            assert str(table_asset.table_name.casefold()) == table_name
            assert isinstance(table_asset.table_name, sqlalchemy.quoted_name)
            assert table_asset.table_name in table_names_in_dbms_schema
            assert table_asset.table_name in quoted_table_names


@pytest.mark.big
@pytest.mark.parametrize(
    ["connection_string", "suggested_datasource_class"],
    [
        ("gregshift://", None),
        ("sqlite:///", "SqliteDatasource"),
        ("snowflake+pyodbc://", "SnowflakeDatasource"),
        ("postgresql+psycopg2://bob:secret@localhost:5432/my_db", "PostgresDatasource"),
        ("${MY_PG_CONN_STR}", "PostgresDatasource"),
        ("databricks://", "DatabricksSQLDatasource"),
    ],
)
def test_specific_datasource_warnings(
    create_engine_fake: None,
    ephemeral_context_with_defaults: EphemeralDataContext,
    monkeypatch: pytest.MonkeyPatch,
    connection_string: str,
    suggested_datasource_class: str | None,
):
    """
    This test ensures that a warning is raised when a specific datasource class is suggested.
    """
    context = ephemeral_context_with_defaults
    monkeypatch.setenv("MY_PG_CONN_STR", "postgresql://bob:secret@localhost:5432/bobs_db")

    if suggested_datasource_class:
        with pytest.warns(GxDatasourceWarning, match=suggested_datasource_class):
            context.data_sources.add_sql(name="my_datasource", connection_string=connection_string)
    else:
        with warnings.catch_warnings():
            context.data_sources.add_sql(
                name="my_datasource", connection_string=connection_string
            ).test_connection()


@pytest.mark.unit
@pytest.mark.parametrize(
    ["input_", "expected_output", "quote_characters"],
    [
        ("my_schema", "my_schema", DEFAULT_QUOTE_CHARACTERS),
        ("MY_SCHEMA", "my_schema", DEFAULT_QUOTE_CHARACTERS),
        ("My_Schema", "my_schema", DEFAULT_QUOTE_CHARACTERS),
        ('"my_schema"', '"my_schema"', DEFAULT_QUOTE_CHARACTERS),
        ('"MY_SCHEMA"', '"MY_SCHEMA"', DEFAULT_QUOTE_CHARACTERS),
        ('"My_Schema"', '"My_Schema"', DEFAULT_QUOTE_CHARACTERS),
        ("'my_schema'", "'my_schema'", DEFAULT_QUOTE_CHARACTERS),
        ("'MY_SCHEMA'", "'MY_SCHEMA'", DEFAULT_QUOTE_CHARACTERS),
        ("'My_Schema'", "'My_Schema'", DEFAULT_QUOTE_CHARACTERS),
        (None, None, DEFAULT_QUOTE_CHARACTERS),
        ("", "", DEFAULT_QUOTE_CHARACTERS),
        ("`My_Schema`", "`My_Schema`", ("`",)),
        ("'My_Schema'", "'my_schema'", ("`",)),
    ],
    ids=lambda x: str(x),
)
def test_to_lower_if_not_quoted(
    input_: str | None, expected_output: str | None, quote_characters: tuple[str, ...]
):
    assert to_lower_if_not_quoted(input_, quote_characters=quote_characters) == expected_output


@pytest.mark.unit
class TestTableAsset:
    @pytest.mark.parametrize("schema_name", ["my_schema", "MY_SCHEMA", "My_Schema"])
    def test_unquoted_schema_names_are_added_as_lowercase(
        self,
        sql_datasource_table_asset_test_connection_noop: SQLDatasource,
        schema_name: str,
    ):
        my_datasource: SQLDatasource = sql_datasource_table_asset_test_connection_noop

        table_asset = my_datasource.add_table_asset(
            name="my_table_asset",
            table_name="my_table",
            schema_name=schema_name,
        )
        assert table_asset.schema_name == schema_name.lower()

    @pytest.mark.parametrize(
        "schema_name",
        [
            '"my_schema"',
            '"MY_SCHEMA"',
            '"My_Schema"',
            "'my_schema'",
            "'MY_SCHEMA'",
            "'My_Schema'",
        ],
    )
    def test_quoted_schema_names_are_not_modified(
        self,
        sql_datasource_table_asset_test_connection_noop: SQLDatasource,
        schema_name: str,
    ):
        my_datasource: SQLDatasource = sql_datasource_table_asset_test_connection_noop

        table_asset = my_datasource.add_table_asset(
            name="my_table_asset",
            table_name="my_table",
            schema_name=schema_name,
        )
        assert table_asset.schema_name == schema_name


if __name__ == "__main__":
    pytest.main([__file__, "-vv"])



from great_expectations.metrics.column import ColumnMetric
from great_expectations.metrics.metric_results import MetricResult


class ColumnValuesMatchRegexValuesResult(MetricResult[list[str]]): ...


class ColumnValuesMatchRegexValues(ColumnMetric[ColumnValuesMatchRegexValuesResult]):
    """List of values in a column that match a regex"""

    name = "column_values.match_regex"
    regex: str
    limit: int = 20



from typing import Sequence

import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.compatibility import pydantic
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.data_sources_and_expectations.test_canonical_expectations import (
    JUST_PANDAS_DATA_SOURCES,
)
from tests.integration.test_utils.data_source_config import (
    DatabricksDatasourceTestConfig,
    DataSourceTestConfig,
    MySQLDatasourceTestConfig,
    PandasDataFrameDatasourceTestConfig,
    PandasFilesystemCsvDatasourceTestConfig,
    PostgreSQLDatasourceTestConfig,
    RedshiftDatasourceTestConfig,
    SnowflakeDatasourceTestConfig,
    SparkFilesystemCsvDatasourceTestConfig,
    SqliteDatasourceTestConfig,
)

COL_A = "col_a"
COL_B = "col_b"


DATA = pd.DataFrame(
    {
        COL_A: ["aa", "ab", "ac", None],
        COL_B: ["aa", "bb", "cc", None],
    }
)

SUPPORTED_DATA_SOURCES: Sequence[DataSourceTestConfig] = [
    PandasDataFrameDatasourceTestConfig(),
    PandasFilesystemCsvDatasourceTestConfig(),
    DatabricksDatasourceTestConfig(),
    MySQLDatasourceTestConfig(),
    PostgreSQLDatasourceTestConfig(),
    RedshiftDatasourceTestConfig(),
    SnowflakeDatasourceTestConfig(),
    SparkFilesystemCsvDatasourceTestConfig(),
    SqliteDatasourceTestConfig(),
]


@parameterize_batch_for_data_sources(data_source_configs=SUPPORTED_DATA_SOURCES, data=DATA)
def test_golden_path(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnValuesToNotMatchRegexList(column=COL_A, regex_list=["x.", "a.."])
    result = batch_for_datasource.validate(expectation)
    assert result.success


class TestNormalSql:
    @pytest.mark.parametrize(
        "expectation",
        [
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchRegexList(column=COL_A, regex_list=["a[x-z]"]),
                id="non_matching",
            ),
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchRegexList(
                    column=COL_A,
                    regex_list=["x.", "a.."],
                ),
                id="multiple_non_matching",
            ),
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchRegexList(
                    column=COL_B, regex_list=["a."], mostly=0.6
                ),
                id="mostly",
            ),
        ],
    )
    @parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
    def test_success(
        self,
        batch_for_datasource: Batch,
        expectation: gxe.ExpectColumnValuesToNotMatchRegexList,
    ) -> None:
        result = batch_for_datasource.validate(expectation)
        assert result.success

    @pytest.mark.parametrize(
        "expectation",
        [
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchRegexList(column=COL_A, regex_list=["^a[abc]$"]),
                id="all_matches",
            ),
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchRegexList(
                    column=COL_A,
                    regex_list=["a.", "x."],
                ),
                id="one_matching",
            ),
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchRegexList(
                    column=COL_A,
                    regex_list=["a.", "\\w+"],
                ),
                id="all_matching",
            ),
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchRegexList(
                    column=COL_B, regex_list=["a."], mostly=0.7
                ),
                id="mostly_threshold_not_met",
            ),
            pytest.param(
                gxe.ExpectColumnValuesToNotMatchRegexList(column=COL_A, regex_list=[""]),
                id="empty_regex",
            ),
        ],
    )
    @parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
    def test_failure(
        self,
        batch_for_datasource: Batch,
        expectation: gxe.ExpectColumnValuesToNotMatchRegexList,
    ) -> None:
        result = batch_for_datasource.validate(expectation)
        assert not result.success


@pytest.mark.unit
def test_invalid_config() -> None:
    with pytest.raises(pydantic.ValidationError):
        gxe.ExpectColumnValuesToNotMatchRegexList(column=COL_A, regex_list=[])



from __future__ import annotations

from great_expectations.compatibility.not_imported import NotImported

PYARROW_NOT_IMPORTED = NotImported("pyarrow is not installed, please 'pip install pyarrow'")

try:
    import pyarrow
except ImportError:
    pyarrow = PYARROW_NOT_IMPORTED



from unittest.mock import Mock  # noqa: TID251 # FIXME CoP

import pytest

from great_expectations.data_context.types.base import (
    ExecutionEngineConfigSchema,
)


@pytest.mark.unit
@pytest.mark.parametrize(
    "connect_args",
    [
        {"connection_factory": Mock()},
        {"ssl_verify_cert": False},
        {"timeout": 30},
        {
            "ssl": {
                "ssl_ca": "ca.pem",
                "ssl_cert": "client-cert.pem",
                "ssl_key": "client-key.pem",
            }
        },
    ],
)
def test_execution_engine_config_conect_args(connect_args):
    """
    this is part of a test-driven fix for: https://github.com/great-expectations/great_expectations/issues/6226
    connect_args examples are here: https://docs.sqlalchemy.org/en/20/core/engines.html#use-the-connect-args-dictionary-parameter
    """
    cfg = ExecutionEngineConfigSchema().load(
        {
            "class_name": "SqlAlchemyExecutionEngine",
            "module_name": "great_expectations.execution_engine",
            "connection_string": "sqlite://",
            "connect_args": connect_args,
        }
    )

    assert cfg.connect_args == connect_args



from tests.integration.backend_dependencies import BackendDependencies
from tests.integration.integration_test_fixture import IntegrationTestFixture

redshift_integration_tests = []

connecting_to_your_data = [
    # TODO: <Alex>ALEX: Rename test modules to include "configured" and "inferred_and_runtime" suffixes in names.</Alex>  # noqa: E501 # FIXME CoP
    # IntegrationTestFixture(
    #     name = "redshift_python_example",
    #     user_flow_script= "docs/docusaurus/docs/snippets/redshift_python_example.py",
    #     data_context_dir= "tests/integration/fixtures/no_datasources/great_expectations",
    #     data_dir= "tests/test_sets/taxi_yellow_tripdata_samples/first_3_files",
    #     backend_dependencies=[ [BackendDependencies.AWS, BackendDependencies.REDSHIFT]],
    #     util_script= "tests/test_utils.py",
    # ),
    # IntegrationTestFixture(
    #     name = "redshift_yaml_example",
    #     user_flow_script= "docs/docusaurus/docs/snippets/redshift_yaml_example.py",
    #     data_context_dir= "tests/integration/fixtures/no_datasources/great_expectations",
    #     data_dir= "tests/test_sets/taxi_yellow_tripdata_samples/first_3_files",
    #     backend_dependencies=[ [BackendDependencies.AWS, BackendDependencies.REDSHIFT]],
    #     util_script= "tests/test_utils.py",
    # ),
]

partition_data = [
    IntegrationTestFixture(
        name="partition_data_on_whole_table_redshift",
        user_flow_script="tests/integration/db/test_sql_data_partitioned_on_whole_table.py",
        data_context_dir="tests/integration/fixtures/no_datasources/great_expectations",
        data_dir="tests/test_sets/taxi_yellow_tripdata_samples/",
        util_script="tests/test_utils.py",
        other_files=(
            (
                "tests/integration/fixtures/partition_and_sample_data/redshift_connection_string.yml",
                "connection_string.yml",
            ),
        ),
        backend_dependencies=[BackendDependencies.REDSHIFT],
    ),
    IntegrationTestFixture(
        name="partition_data_on_datetime_redshift",
        user_flow_script="tests/integration/db/test_sql_data_partitioned_on_datetime_and_day_part.py",
        data_context_dir="tests/integration/fixtures/no_datasources/great_expectations",
        data_dir="tests/test_sets/taxi_yellow_tripdata_samples/",
        util_script="tests/test_utils.py",
        other_files=(
            (
                "tests/integration/fixtures/partition_and_sample_data/redshift_connection_string.yml",
                "connection_string.yml",
            ),
        ),
        backend_dependencies=[BackendDependencies.REDSHIFT],
    ),
]

sample_data = []

deployment_patterns = [
    IntegrationTestFixture(
        name="deployment_patterns_redshift",
        user_flow_script="docs/docusaurus/docs/snippets/aws_redshift_deployment_patterns.py",
        data_context_dir="tests/integration/fixtures/no_datasources/great_expectations",
        data_dir="tests/test_sets/taxi_yellow_tripdata_samples/",
        backend_dependencies=[BackendDependencies.REDSHIFT],
    ),
]

redshift_integration_tests += connecting_to_your_data
redshift_integration_tests += partition_data
redshift_integration_tests += sample_data
redshift_integration_tests += deployment_patterns



from __future__ import annotations

import pytest

from great_expectations.data_context.store.data_asset_store import DataAssetStore


@pytest.mark.cloud
@pytest.mark.parametrize(
    "response_json, expected, error_type",
    [
        pytest.param(
            {
                "data": {
                    "id": "03d61d4e-003f-48e7-a3b2-f9f842384da3",
                    "attributes": {
                        "data_asset_config": {
                            "name": "my_asset",
                            "type": "pandas",
                        },
                    },
                }
            },
            {
                "id": "03d61d4e-003f-48e7-a3b2-f9f842384da3",
                "name": "my_asset",
                "type": "pandas",
            },
            None,
            id="single_config",
        ),
        pytest.param(
            {
                "data": [
                    {
                        "id": "03d61d4e-003f-48e7-a3b2-f9f842384da3",
                        "attributes": {
                            "data_asset_config": {
                                "name": "my_asset",
                                "type": "pandas",
                            },
                        },
                    }
                ]
            },
            {
                "id": "03d61d4e-003f-48e7-a3b2-f9f842384da3",
                "name": "my_asset",
                "type": "pandas",
            },
            None,
            id="single_config_in_list",
        ),
        pytest.param(
            {
                "data": [
                    {
                        "data": [
                            {
                                "id": "03d61d4e-003f-48e7-a3b2-f9f842384da3",
                                "attributes": {
                                    "data_asset_config": {
                                        "name": "my_asset",
                                        "type": "pandas",
                                    },
                                },
                            }
                        ]
                    },
                    {
                        "data": [
                            {
                                "id": "ffg61d4e-003f-48e7-a3b2-f9f842384da3",
                                "attributes": {
                                    "data_asset_config": {
                                        "name": "my_other_asset",
                                        "type": "pandas",
                                    },
                                },
                            }
                        ]
                    },
                ]
            },
            None,
            TypeError,
            id="multiple_config_in_list",
        ),
    ],
)
def test_gx_cloud_response_json_to_object_dict(
    response_json: dict, expected: dict | None, error_type: Exception | None
) -> None:
    if error_type:
        with pytest.raises(error_type):
            _ = DataAssetStore.gx_cloud_response_json_to_object_dict(response_json)
    else:
        actual = DataAssetStore.gx_cloud_response_json_to_object_dict(response_json)
        assert actual == expected



import pathlib
import shutil
from unittest import mock

import pytest

import great_expectations as gx
from great_expectations.data_context import CloudDataContext, EphemeralDataContext
from great_expectations.data_context.cloud_constants import GXCloudEnvironmentVariable
from great_expectations.data_context.data_context.file_data_context import (
    FileDataContext,
)
from great_expectations.data_context.types.base import (
    DataContextConfig,
    InMemoryStoreBackendDefaults,
)
from great_expectations.exceptions.exceptions import (
    GitIgnoreScaffoldingError,
    GXCloudConfigurationError,
)
from tests.test_utils import working_directory

GX_CLOUD_PARAMS_ALL = {
    "cloud_base_url": "localhost:7000",
    "cloud_organization_id": "bd20fead-2c31-4392-bcd1-f1e87ad5a79c",
    "cloud_access_token": "i_am_a_token",
}
GX_CLOUD_PARAMS_REQUIRED = {
    "cloud_organization_id": "bd20fead-2c31-4392-bcd1-f1e87ad5a79c",
    "cloud_access_token": "i_am_a_token",
}


@pytest.fixture()
def set_up_cloud_envs(monkeypatch):
    monkeypatch.setenv("GX_CLOUD_BASE_URL", "localhost:7000")
    monkeypatch.setenv("GX_CLOUD_ORGANIZATION_ID", "bd20fead-2c31-4392-bcd1-f1e87ad5a79c")
    monkeypatch.setenv("GX_CLOUD_ACCESS_TOKEN", "i_am_a_token")


@pytest.fixture
def clear_env_vars(monkeypatch):
    # Delete local env vars (if present)
    for env_var in GXCloudEnvironmentVariable:
        monkeypatch.delenv(env_var, raising=False)


@pytest.mark.unit
def test_base_context(clear_env_vars):
    config: DataContextConfig = DataContextConfig(
        config_version=3.0,
        plugins_directory=None,
        expectations_store_name="expectations_store",
        checkpoint_store_name="checkpoint_store",
        stores={
            "expectations_store": {"class_name": "ExpectationsStore"},
            "checkpoint_store": {"class_name": "CheckpointStore"},
            "validation_result_store": {"class_name": "ValidationResultsStore"},
            "validation_definition_store": {"class_name": "ValidationDefinitionStore"},
        },
        validation_results_store_name="validation_result_store",
        data_docs_sites={},
    )
    assert isinstance(gx.get_context(project_config=config), EphemeralDataContext)


@pytest.mark.unit
def test_base_context__with_overridden_yml(tmp_path: pathlib.Path, clear_env_vars):
    project_path = tmp_path / "empty_data_context"
    project_path.mkdir()
    context_path = project_path / FileDataContext.GX_DIR
    context = gx.get_context(context_root_dir=context_path)
    assert isinstance(context, FileDataContext)
    assert context.expectations_store_name == "expectations_store"

    config: DataContextConfig = DataContextConfig(
        config_version=3.0,
        plugins_directory=None,
        expectations_store_name="new_expectations_store",
        checkpoint_store_name="new_checkpoint_store",
        stores={
            "new_expectations_store": {"class_name": "ExpectationsStore"},
            "new_checkpoint_store": {"class_name": "CheckpointStore"},
            "new_validation_result_store": {"class_name": "ValidationResultsStore"},
        },
        validation_results_store_name="new_validation_result_store",
        data_docs_sites={},
    )
    context = gx.get_context(project_config=config, context_root_dir=context_path)
    assert isinstance(context, FileDataContext)
    assert context.expectations_store_name == "new_expectations_store"


@pytest.mark.unit
def test_data_context_root_dir_returns_data_context(
    tmp_path: pathlib.Path,
    clear_env_vars,
):
    project_path = tmp_path / "empty_data_context"
    project_path.mkdir()
    context_path = project_path / FileDataContext.GX_DIR
    assert isinstance(gx.get_context(context_root_dir=str(context_path)), FileDataContext)


@pytest.mark.unit
def test_base_context_invalid_root_dir(clear_env_vars, tmp_path):
    config: DataContextConfig = DataContextConfig(
        config_version=3.0,
        plugins_directory=None,
        expectations_store_name="expectations_store",
        checkpoint_store_name="checkpoint_store",
        stores={
            "expectations_store": {"class_name": "ExpectationsStore"},
            "checkpoint_store": {"class_name": "CheckpointStore"},
            "validation_result_store": {"class_name": "ValidationResultsStore"},
        },
        validation_results_store_name="validation_result_store",
        data_docs_sites={},
    )

    context_root_dir = tmp_path / "root"
    context_root_dir.mkdir()
    assert isinstance(
        gx.get_context(project_config=config, context_root_dir=context_root_dir),
        FileDataContext,
    )


@pytest.mark.parametrize("ge_cloud_mode", [True, None])
@pytest.mark.cloud
def test_cloud_context_env(set_up_cloud_envs, empty_ge_cloud_data_context_config, ge_cloud_mode):
    with mock.patch.object(
        CloudDataContext,
        "retrieve_data_context_config_from_cloud",
        return_value=empty_ge_cloud_data_context_config,
    ):
        assert isinstance(
            gx.get_context(cloud_mode=ge_cloud_mode),
            CloudDataContext,
        )


@pytest.mark.cloud
def test_cloud_missing_env_throws_exception(clear_env_vars, empty_ge_cloud_data_context_config):
    with pytest.raises(GXCloudConfigurationError):
        gx.get_context(cloud_mode=True)


@pytest.mark.parametrize("params", [GX_CLOUD_PARAMS_REQUIRED, GX_CLOUD_PARAMS_ALL])
@pytest.mark.cloud
def test_cloud_context_params(monkeypatch, empty_ge_cloud_data_context_config, params):
    with mock.patch.object(
        CloudDataContext,
        "retrieve_data_context_config_from_cloud",
        return_value=empty_ge_cloud_data_context_config,
    ):
        assert isinstance(
            gx.get_context(**params),
            CloudDataContext,
        )


@pytest.mark.cloud
def test_cloud_context_with_in_memory_config_overrides(
    monkeypatch, empty_ge_cloud_data_context_config
):
    with mock.patch.object(
        CloudDataContext,
        "retrieve_data_context_config_from_cloud",
        return_value=empty_ge_cloud_data_context_config,
    ):
        context = gx.get_context(
            cloud_base_url="localhost:7000",
            cloud_organization_id="bd20fead-2c31-4392-bcd1-f1e87ad5a79c",
            cloud_access_token="i_am_a_token",
        )
        assert isinstance(context, CloudDataContext)
        assert context.expectations_store_name == "default_expectations_store"

        config: DataContextConfig = DataContextConfig(
            config_version=3.0,
            plugins_directory=None,
            expectations_store_name="new_expectations_store",
            checkpoint_store_name="new_checkpoint_store",
            stores={
                "new_expectations_store": {"class_name": "ExpectationsStore"},
                "new_checkpoint_store": {"class_name": "CheckpointStore"},
                "new_validation_result_store": {"class_name": "ValidationResultsStore"},
            },
            validation_results_store_name="new_validation_result_store",
            data_docs_sites={},
        )
        context = gx.get_context(
            project_config=config,
            cloud_base_url="localhost:7000",
            cloud_organization_id="bd20fead-2c31-4392-bcd1-f1e87ad5a79c",
            cloud_access_token="i_am_a_token",
        )
        assert isinstance(context, CloudDataContext)
        assert context.expectations_store_name == "new_expectations_store"


@pytest.mark.unit
def test_get_context_with_no_arguments_returns_ephemeral_with_sensible_defaults():
    context = gx.get_context()
    assert isinstance(context, EphemeralDataContext)

    defaults = InMemoryStoreBackendDefaults(init_temp_docs_sites=True)
    assert context.config.stores == defaults.stores


@pytest.mark.unit
def test_get_context_with_mode_equals_ephemeral_returns_ephemeral_data_context():
    context = gx.get_context(mode="ephemeral")
    assert isinstance(context, EphemeralDataContext)


@pytest.mark.unit
def test_get_context_with_mode_equals_file_returns_file_data_context(
    tmp_path: pathlib.Path,
):
    with working_directory(tmp_path):
        context = gx.get_context(mode="file")
    assert isinstance(context, FileDataContext)


@pytest.mark.cloud
def test_get_context_with_mode_equals_cloud_returns_cloud_data_context(
    empty_ge_cloud_data_context_config: DataContextConfig, set_up_cloud_envs
):
    with mock.patch.object(
        CloudDataContext,
        "retrieve_data_context_config_from_cloud",
        return_value=empty_ge_cloud_data_context_config,
    ) as mock_retrieve_config:
        context = gx.get_context(mode="cloud")

    mock_retrieve_config.assert_called_once()
    assert isinstance(context, CloudDataContext)


@pytest.mark.filesystem
def test_get_context_with_context_root_dir_scaffolds_filesystem(tmp_path: pathlib.Path):
    root = tmp_path / "root"
    context_root_dir = root.joinpath(FileDataContext.GX_DIR)
    assert not context_root_dir.exists()

    context = gx.get_context(context_root_dir=context_root_dir)

    assert isinstance(context, FileDataContext)
    assert context_root_dir.exists()
    assert (context_root_dir / FileDataContext.GITIGNORE).read_text() == "\nuncommitted/"


@pytest.mark.filesystem
def test_get_context_with_custom_context_root_dir_scaffolds_filesystem(tmp_path: pathlib.Path):
    root = tmp_path / "root"
    context_root_dir = root.joinpath("hello_world")
    assert not context_root_dir.exists()

    context = gx.get_context(context_root_dir=context_root_dir)

    assert isinstance(context, FileDataContext)
    assert context_root_dir.exists()
    assert (context_root_dir / FileDataContext.GITIGNORE).read_text() == "\nuncommitted/"


@pytest.mark.filesystem
def test_get_context_with_mode_and_custom_context_root_dir_scaffolds_filesystem(
    tmp_path: pathlib.Path,
):
    root = tmp_path / "root"
    context_root_dir = root.joinpath("hello_world")
    assert not context_root_dir.exists()

    context = gx.get_context(mode="file", context_root_dir=context_root_dir)

    assert isinstance(context, FileDataContext)
    assert context_root_dir.exists()
    assert (context_root_dir / FileDataContext.GITIGNORE).read_text() == "\nuncommitted/"


@pytest.mark.filesystem
def test_errors_if_context_root_dir_and_project_root_dir_are_both_provided_for_file_context(
    tmp_path: pathlib.Path,
):
    root = tmp_path / "root"
    context_root_dir = root.joinpath("hello_world")
    assert not context_root_dir.exists()

    with pytest.raises(
        TypeError,
        match="'project_root_dir' and 'context_root_dir' are conflicting args; please only provide one",  # noqa: E501
    ):
        gx.get_context(  # type: ignore[call-overload]
            mode="file",
            context_root_dir=context_root_dir,
            project_root_dir=context_root_dir.parent,
        )


@pytest.mark.filesystem
def test_get_context_with_context_root_dir_scaffolds_existing_gitignore(clear_env_vars, tmp_path):
    context_root_dir = tmp_path / FileDataContext.GX_DIR
    context_root_dir.mkdir()
    with open(context_root_dir / FileDataContext.GITIGNORE, "w") as f:
        f.write("asdf")

    context = gx.get_context(context_root_dir=context_root_dir)

    assert isinstance(context, FileDataContext)
    assert (context_root_dir / FileDataContext.GITIGNORE).read_text() == "asdf\nuncommitted/"


@pytest.mark.filesystem
def test_get_context_with_context_root_dir_scaffolds_new_gitignore(clear_env_vars, tmp_path):
    context_root_dir = tmp_path / FileDataContext.GX_DIR
    context_root_dir.mkdir()

    context = gx.get_context(context_root_dir=context_root_dir)

    assert isinstance(context, FileDataContext)
    assert (context_root_dir / FileDataContext.GITIGNORE).read_text() == "\nuncommitted/"


@pytest.mark.filesystem
def test_get_context_with_context_root_dir_gitignore_error(clear_env_vars, tmp_path):
    context_root_dir = tmp_path / FileDataContext.GX_DIR
    context_root_dir.mkdir()

    with mock.patch(
        "great_expectations.data_context.data_context.serializable_data_context.SerializableDataContext._scaffold_gitignore",
        side_effect=OSError("Error"),
    ):
        with pytest.raises(GitIgnoreScaffoldingError):
            gx.get_context(context_root_dir=context_root_dir)


@pytest.mark.filesystem
def test_get_context_scaffolds_gx_dir(tmp_path: pathlib.Path):
    with working_directory(tmp_path):
        context = gx.get_context(mode="file")
    assert isinstance(context, FileDataContext)

    project_root_dir = pathlib.Path(context.root_directory)
    assert project_root_dir.stem == FileDataContext.GX_DIR


@pytest.mark.filesystem
def test_get_context_finds_legacy_great_expectations_dir(
    tmp_path: pathlib.Path,
):
    working_dir = tmp_path / "a" / "b" / "c" / "d" / "working_dir"

    # Scaffold great_expectations
    context_root_dir = working_dir / FileDataContext._LEGACY_GX_DIR
    context_root_dir.mkdir(parents=True)

    # Scaffold great_expectations.yml
    gx_yml = context_root_dir / FileDataContext.GX_YML
    yml_fixture = (
        pathlib.Path(__file__)
        .joinpath("../../test_fixtures/great_expectations_basic.yml")
        .resolve()
    )
    assert yml_fixture.exists()
    shutil.copy(yml_fixture, gx_yml)

    with working_directory(working_dir):
        context = gx.get_context()
    assert isinstance(context, FileDataContext)

    project_root_dir = pathlib.Path(context.root_directory)
    assert project_root_dir.stem == FileDataContext._LEGACY_GX_DIR



from __future__ import annotations

import datetime
import json
import uuid
from typing import TYPE_CHECKING, Type
from unittest import mock
from unittest.mock import ANY

import pandas as pd
import pytest

import great_expectations as gx
import great_expectations.expectations as gxe
from great_expectations import RunIdentifier
from great_expectations import __version__ as GX_VERSION
from great_expectations.core.batch_definition import BatchDefinition
from great_expectations.core.expectation_suite import ExpectationSuite
from great_expectations.core.expectation_validation_result import (
    ExpectationSuiteValidationResult,
    ExpectationValidationResult,
)
from great_expectations.core.result_format import ResultFormat
from great_expectations.core.validation_definition import ValidationDefinition
from great_expectations.data_context.data_context.abstract_data_context import AbstractDataContext
from great_expectations.data_context.data_context.cloud_data_context import (
    CloudDataContext,
)
from great_expectations.data_context.data_context.context_factory import (
    ProjectManager,
    set_context,
)
from great_expectations.data_context.data_context.ephemeral_data_context import (
    EphemeralDataContext,
)
from great_expectations.data_context.store.validation_results_store import ValidationResultsStore
from great_expectations.data_context.types.refs import GXCloudResourceRef
from great_expectations.data_context.types.resource_identifiers import (
    GXCloudIdentifier,
    ValidationResultIdentifier,
)
from great_expectations.datasource.fluent.pandas_datasource import (
    CSVAsset,
    PandasDatasource,
    _PandasDataAsset,
)
from great_expectations.exceptions import (
    BatchDefinitionNotAddedError,
    ExpectationSuiteNotAddedError,
    ResourceFreshnessError,
    ValidationDefinitionNotAddedError,
    ValidationDefinitionRelatedResourcesFreshnessError,
)
from great_expectations.exceptions.exceptions import (
    BatchDefinitionNotFoundError,
    ExpectationSuiteNotFoundError,
    ValidationDefinitionNotFoundError,
)
from great_expectations.exceptions.resource_freshness import ResourceFreshnessAggregateError
from great_expectations.execution_engine.execution_engine import ExecutionEngine
from great_expectations.expectations.expectation_configuration import (
    ExpectationConfiguration,
)
from great_expectations.validator.v1_validator import (
    OldValidator,
)

if TYPE_CHECKING:
    from unittest.mock import MagicMock  # noqa: TID251 # FIXME CoP

    from pytest_mock import MockerFixture

BATCH_ID = "my_batch_id"
DATA_SOURCE_NAME = "my_datasource"
ASSET_NAME = "csv_asset"
BATCH_DEFINITION_NAME = "my_batch_definition"
ACTIVE_BATCH_SPEC = {
    "type": "table",
    "data_asset_name": ASSET_NAME,
    "table_name": "test_table",
    "schema_name": "test_schema",
    "batch_identifiers": {"date": {"year": 2017, "month": 12, "day": 3}},
    "partitioner_method": "partition_on_year_and_month_and_day",
    "partitioner_kwargs": {"column_name": "date"},
}
ACTIVE_BATCH_DEFINITION = {
    "datasource_name": DATA_SOURCE_NAME,
    "data_connector_name": "fluent",
    "data_asset_name": ASSET_NAME,
    "batch_identifiers": {"date": {"year": 2017, "month": 12, "day": 3}},
}
BATCH_MARKERS = {"ge_load_time": "20240814T172846.050804Z"}


@pytest.fixture
def ephemeral_context():
    return gx.get_context(mode="ephemeral")


@pytest.fixture
def validation_definition(ephemeral_context: EphemeralDataContext) -> ValidationDefinition:
    context = ephemeral_context
    batch_definition = (
        context.data_sources.add_pandas(DATA_SOURCE_NAME)
        .add_csv_asset(ASSET_NAME, "taxi.csv")  # type: ignore # FIXME CoP
        .add_batch_definition(BATCH_DEFINITION_NAME)
    )
    return context.validation_definitions.add(
        ValidationDefinition(
            name="my_validation",
            data=batch_definition,
            suite=context.suites.add(ExpectationSuite(name="my_suite")),
        )
    )


@pytest.fixture
def dataframe_validation_definition(
    ephemeral_context: EphemeralDataContext,
) -> ValidationDefinition:
    context = ephemeral_context
    batch_definition = (
        context.data_sources.add_pandas(DATA_SOURCE_NAME)
        .add_dataframe_asset("dataframe_asset")
        .add_batch_definition_whole_dataframe("dataframe_batch_def")
    )
    return context.validation_definitions.add(
        ValidationDefinition(
            name="my_dataframe_validation",
            data=batch_definition,
            suite=context.suites.add(ExpectationSuite(name="my_suite")),
        )
    )


@pytest.fixture
def postgres_validation_definition(
    ephemeral_context: EphemeralDataContext,
) -> ValidationDefinition:
    context = ephemeral_context
    batch_definition = (
        ephemeral_context.data_sources.add_postgres(
            name="postgres_datasource",
            connection_string="postgresql+psycopg2://postgres:postgres@localhost:5432/test_ci",
        )
        .add_query_asset(name="my_asset", query="SELECT * FROM fake_table")
        .add_batch_definition_monthly(
            name="my_batch_definition",
            column="not_very_real",
            # We skip validating we can generate batches because fake_table is fake
            # and won't validate.
            validate_batchable=False,
        )
    )
    return context.validation_definitions.add(
        ValidationDefinition(
            name="my_postgres_validation",
            data=batch_definition,
            suite=context.suites.add(ExpectationSuite(name="my_suite")),
        )
    )


@pytest.fixture
def cloud_validation_definition(
    empty_cloud_data_context: CloudDataContext,
) -> ValidationDefinition:
    context = empty_cloud_data_context
    batch_definition = (
        empty_cloud_data_context.data_sources.add_pandas(DATA_SOURCE_NAME)
        .add_csv_asset(ASSET_NAME, "taxi.csv")  # type: ignore # FIXME CoP
        .add_batch_definition(BATCH_DEFINITION_NAME)
    )
    suite = context.suites.add(ExpectationSuite(name="my_suite"))
    return context.validation_definitions.add(
        ValidationDefinition(
            name="my_validation",
            data=batch_definition,
            suite=suite,
        )
    )


@pytest.mark.unit
def test_validation_definition_data_properties(validation_definition: ValidationDefinition):
    assert validation_definition.data.name == BATCH_DEFINITION_NAME
    assert validation_definition.batch_definition.name == BATCH_DEFINITION_NAME
    assert validation_definition.asset.name == ASSET_NAME
    assert validation_definition.data_source.name == DATA_SOURCE_NAME


class TestValidationRun:
    @pytest.fixture
    def mock_validator(self, mocker: MockerFixture):
        """Set up our ProjectManager to return a mock Validator"""
        with mock.patch.object(ProjectManager, "get_validator") as mock_get_validator:
            with mock.patch.object(OldValidator, "graph_validate"):
                gx.get_context(mode="ephemeral")
                mock_execution_engine = mocker.MagicMock(
                    spec=ExecutionEngine,
                    batch_manager=mocker.MagicMock(
                        active_batch_id=BATCH_ID,
                        active_batch_spec=ACTIVE_BATCH_SPEC,
                        active_batch_definition=ACTIVE_BATCH_DEFINITION,
                        active_batch_markers=BATCH_MARKERS,
                    ),
                )
                mock_validator = OldValidator(execution_engine=mock_execution_engine)
                mock_get_validator.return_value = mock_validator

                yield mock_validator

    @pytest.mark.unit
    def test_passes_simple_data_to_validator(
        self,
        mock_validator: MagicMock,
        validation_definition: ValidationDefinition,
    ):
        validation_definition.suite.add_expectation(
            gxe.ExpectColumnMaxToBeBetween(column="foo", max_value=1)
        )
        mock_validator.graph_validate.return_value = [ExpectationValidationResult(success=True)]

        validation_definition.run()

        mock_validator.graph_validate.assert_called_with(
            configurations=[
                ExpectationConfiguration(
                    type="expect_column_max_to_be_between",
                    kwargs={"column": "foo", "max_value": 1.0},
                )
            ],
            runtime_configuration={"result_format": "SUMMARY"},
        )

    @mock.patch.object(_PandasDataAsset, "build_batch_request", autospec=True)
    @pytest.mark.unit
    def test_passes_complex_data_to_validator(
        self,
        mock_build_batch_request,
        mock_validator: MagicMock,
        validation_definition: ValidationDefinition,
    ):
        validation_definition.suite.add_expectation(
            gxe.ExpectColumnMaxToBeBetween(column="foo", max_value={"$PARAMETER": "max_value"})
        )
        mock_validator.graph_validate.return_value = [ExpectationValidationResult(success=True)]

        validation_definition.run(
            batch_parameters={"year": 2024},
            expectation_parameters={"max_value": 9000},
            result_format=ResultFormat.COMPLETE,
        )

        mock_validator.graph_validate.assert_called_with(
            configurations=[
                ExpectationConfiguration(
                    type="expect_column_max_to_be_between",
                    kwargs={"column": "foo", "max_value": 9000},
                )
            ],
            runtime_configuration={"result_format": "COMPLETE"},
        )

    @pytest.mark.unit
    def test_returns_expected_data(
        self,
        mock_validator: MagicMock,
        validation_definition: ValidationDefinition,
    ):
        graph_validate_results = [ExpectationValidationResult(success=True)]
        mock_validator.graph_validate.return_value = graph_validate_results

        output = validation_definition.run()

        # Ignore meta for purposes of this test
        output["meta"] = {}
        assert output == ExpectationSuiteValidationResult(
            results=graph_validate_results,
            success=True,
            suite_name="empty_suite",
            statistics={
                "evaluated_expectations": 1,
                "successful_expectations": 1,
                "unsuccessful_expectations": 0,
                "success_percent": 100.0,
            },
            meta={},
        )

    @pytest.mark.parametrize("checkpoint_id", [None, "my_checkpoint_id"])
    @pytest.mark.unit
    def test_adds_requisite_fields(
        self,
        mock_validator: MagicMock,
        validation_definition: ValidationDefinition,
        checkpoint_id: str | None,
    ):
        mock_validator.graph_validate.return_value = []

        output = validation_definition.run(checkpoint_id=checkpoint_id)

        assert output.meta == {
            "validation_id": validation_definition.id,
            "checkpoint_id": checkpoint_id,
            "batch_parameters": None,
            "batch_spec": ACTIVE_BATCH_SPEC,
            "batch_markers": BATCH_MARKERS,
            "active_batch_definition": ACTIVE_BATCH_DEFINITION,
            "great_expectations_version": GX_VERSION,
            "run_id": ANY,
            "validation_time": ANY,
        }
        assert isinstance(output.meta["run_id"], RunIdentifier)
        assert isinstance(output.meta["validation_time"], datetime.datetime)

    @pytest.mark.unit
    def test_adds_correct_batch_parameter_field_for_dataframes(
        self,
        mock_validator: MagicMock,
        dataframe_validation_definition: ValidationDefinition,
    ) -> None:
        mock_validator.graph_validate.return_value = []

        output = dataframe_validation_definition.run(
            checkpoint_id=None,
            batch_parameters={"dataframe": pd.DataFrame({"a": ["1", "2", "3", "4", "5"]})},
        )

        assert output.meta == {
            "validation_id": dataframe_validation_definition.id,
            "checkpoint_id": None,
            "batch_parameters": {"dataframe": "<DATAFRAME>"},
            "batch_spec": ACTIVE_BATCH_SPEC,
            "batch_markers": BATCH_MARKERS,
            "active_batch_definition": ACTIVE_BATCH_DEFINITION,
            "great_expectations_version": GX_VERSION,
            "run_id": ANY,
            "validation_time": ANY,
        }
        assert isinstance(output.meta["run_id"], RunIdentifier)
        assert isinstance(output.meta["validation_time"], datetime.datetime)

    @pytest.mark.parametrize(
        "batch_parameters",
        [
            pytest.param(None),
            pytest.param({"year": 2024}),
            pytest.param({"year": 2024, "month": 10}),
        ],
    )
    @pytest.mark.postgresql
    def test_adds_correct_batch_parameter_fields_for_postgres(
        self,
        mock_validator: MagicMock,
        postgres_validation_definition: ValidationDefinition,
        batch_parameters: dict | None,
    ) -> None:
        mock_validator.graph_validate.return_value = []

        output = postgres_validation_definition.run(
            checkpoint_id=None,
            batch_parameters=batch_parameters,
        )

        assert output.meta == {
            "validation_id": postgres_validation_definition.id,
            "checkpoint_id": None,
            "batch_parameters": batch_parameters,
            "batch_spec": ACTIVE_BATCH_SPEC,
            "batch_markers": BATCH_MARKERS,
            "active_batch_definition": ACTIVE_BATCH_DEFINITION,
            "great_expectations_version": GX_VERSION,
            "run_id": ANY,
            "validation_time": ANY,
        }
        assert isinstance(output.meta["run_id"], RunIdentifier)
        assert isinstance(output.meta["validation_time"], datetime.datetime)

    @mock.patch.object(ValidationResultsStore, "set")
    @pytest.mark.unit
    def test_persists_validation_results_for_non_cloud(
        self,
        mock_validation_results_store_set: MagicMock,
        mock_validator: MagicMock,
        validation_definition: ValidationDefinition,
    ):
        validation_definition.suite.add_expectation(
            gxe.ExpectColumnMaxToBeBetween(column="foo", max_value=1)
        )
        mock_validator.graph_validate.return_value = [ExpectationValidationResult(success=True)]

        validation_definition.run()

        mock_validator.graph_validate.assert_called_with(
            configurations=[
                ExpectationConfiguration(
                    type="expect_column_max_to_be_between",
                    kwargs={"column": "foo", "max_value": 1.0},
                )
            ],
            runtime_configuration={"result_format": "SUMMARY"},
        )

        # validate we are calling set on the store with data that's roughly the right shape
        [(_, kwargs)] = mock_validation_results_store_set.call_args_list
        key = kwargs["key"]
        value = kwargs["value"]
        assert isinstance(key, ValidationResultIdentifier)
        assert key.batch_identifier == BATCH_ID
        assert value.success is True

    @mock.patch.object(ValidationResultsStore, "set")
    @pytest.mark.unit
    def test_persists_validation_results_for_cloud(
        self,
        mock_validation_results_store_set: MagicMock,
        mock_validator: MagicMock,
        cloud_validation_definition: ValidationDefinition,
    ):
        expectation = gxe.ExpectColumnMaxToBeBetween(column="foo", max_value=1)
        cloud_validation_definition.suite.add_expectation(expectation=expectation)
        cloud_validation_definition.suite.save()
        mock_validator.graph_validate.return_value = [
            ExpectationValidationResult(success=True, expectation_config=expectation.configuration)
        ]

        cloud_validation_definition.run()

        # validate we are calling set on the store with data that's roughly the right shape
        [(_, kwargs)] = mock_validation_results_store_set.call_args_list
        key = kwargs["key"]
        value = kwargs["value"]
        assert isinstance(key, GXCloudIdentifier)
        assert value.success is True

    @mock.patch.object(
        ValidationResultsStore,
        "set",
        return_value=GXCloudResourceRef(
            resource_type="validation_result",
            id="59b72ca5-4636-44be-a367-46b54ae51fe1",
            url="https://api.greatexpectations.io/api/v1/organizations/11111111-ba69-4295-8fe1-61eef96f12b4/validation-results",
            response_json={"data": {"result_url": "my_result_url"}},
        ),
    )
    @pytest.mark.unit
    def test_cloud_validation_def_adds_id_and_url_to_result(
        self,
        mock_validation_results_store_set: MagicMock,
        mock_validator: MagicMock,
        cloud_validation_definition: ValidationDefinition,
    ):
        expectation = gxe.ExpectColumnMaxToBeBetween(column="foo", max_value=1)
        cloud_validation_definition.suite.add_expectation(expectation=expectation)
        cloud_validation_definition.suite.save()
        mock_validator.graph_validate.return_value = [
            ExpectationValidationResult(success=True, expectation_config=expectation.configuration)
        ]

        result = cloud_validation_definition.run()

        assert result.id == "59b72ca5-4636-44be-a367-46b54ae51fe1"
        assert result.result_url == "my_result_url"

    @mock.patch.object(ValidationResultsStore, "set")
    @pytest.mark.unit
    def test_cloud_validation_def_creates_rendered_content(
        self,
        mock_validation_results_store_set: MagicMock,
        mock_validator: MagicMock,
        cloud_validation_definition: ValidationDefinition,
    ):
        expectation = gxe.ExpectColumnMaxToBeBetween(column="foo", max_value=1)
        cloud_validation_definition.suite.add_expectation(expectation=expectation)
        cloud_validation_definition.suite.save()
        mock_validator.graph_validate.return_value = [
            ExpectationValidationResult(success=True, expectation_config=expectation.configuration)
        ]

        result = cloud_validation_definition.run()

        assert len(result.results) == 1
        assert result.results[0].expectation_config is not None
        assert result.results[0].expectation_config.rendered_content is not None
        assert result.results[0].rendered_content is not None

    @pytest.mark.unit
    def test_dependencies_not_added_raises_error(self, validation_definition: ValidationDefinition):
        validation_definition.suite.id = None
        validation_definition.data.id = None

        with pytest.raises(ValidationDefinitionRelatedResourcesFreshnessError) as e:
            validation_definition.run()

        assert [type(err) for err in e.value.errors] == [
            BatchDefinitionNotAddedError,
            ExpectationSuiteNotAddedError,
        ]


class TestValidationDefinitionSerialization:
    ds_name = "my_ds"
    asset_name = "my_asset"
    batch_definition_name = "my_batch_definition"
    suite_name = "my_suite"
    validation_definition_name = "my_validation"

    @pytest.fixture
    def context(self, in_memory_runtime_context: EphemeralDataContext) -> EphemeralDataContext:
        return in_memory_runtime_context

    @pytest.fixture
    def validation_definition_data(
        self,
        context: EphemeralDataContext,
    ) -> tuple[PandasDatasource, CSVAsset, BatchDefinition]:
        ds = context.data_sources.add_pandas(self.ds_name)
        asset = ds.add_csv_asset(self.asset_name, "data.csv")
        batch_definition = asset.add_batch_definition(self.batch_definition_name)

        return ds, asset, batch_definition

    @pytest.fixture
    def validation_definition_suite(self, context: EphemeralDataContext) -> ExpectationSuite:
        return context.suites.add(ExpectationSuite(self.suite_name))

    @pytest.mark.unit
    def test_validation_definition_serialization(
        self,
        in_memory_runtime_context: EphemeralDataContext,
        validation_definition_data: tuple[PandasDatasource, CSVAsset, BatchDefinition],
        validation_definition_suite: ExpectationSuite,
    ):
        context = in_memory_runtime_context
        pandas_ds, csv_asset, batch_definition = validation_definition_data

        ds_id = str(uuid.uuid4())
        pandas_ds.id = ds_id

        asset_id = str(uuid.uuid4())
        csv_asset.id = asset_id

        batch_definition_id = str(uuid.uuid4())
        batch_definition.id = batch_definition_id

        suite_id = str(uuid.uuid4())
        validation_definition_suite.id = suite_id

        validation_definition = context.validation_definitions.add(
            ValidationDefinition(
                name=self.validation_definition_name,
                data=batch_definition,
                suite=validation_definition_suite,
            )
        )

        actual = json.loads(validation_definition.json(models_as_dict=False))
        expected = {
            "name": self.validation_definition_name,
            "data": {
                "datasource": {
                    "name": pandas_ds.name,
                    "id": ds_id,
                },
                "asset": {
                    "name": csv_asset.name,
                    "id": asset_id,
                },
                "batch_definition": {
                    "name": batch_definition.name,
                    "id": batch_definition_id,
                },
            },
            "suite": {
                "name": validation_definition_suite.name,
                "id": suite_id,
            },
            "id": mock.ANY,
        }

        assert actual == expected
        assert actual["id"] is not None

    def _assert_contains_valid_uuid(self, data: dict):
        id = data.pop("id")
        data["id"] = mock.ANY
        try:
            uuid.UUID(id)
        except ValueError:
            pytest.fail(f"Expected {id} to be a valid UUID")

    @pytest.mark.unit
    def test_validation_definition_deserialization_success(
        self,
        context: EphemeralDataContext,
        validation_definition_data: tuple[PandasDatasource, CSVAsset, BatchDefinition],
        validation_definition_suite: ExpectationSuite,
    ):
        _, _, batch_definition = validation_definition_data

        serialized_config = {
            "name": self.validation_definition_name,
            "data": {
                "datasource": {
                    "name": self.ds_name,
                    "id": None,
                },
                "asset": {
                    "name": self.asset_name,
                    "id": None,
                },
                "batch_definition": {
                    "name": self.batch_definition_name,
                    "id": None,
                },
            },
            "suite": {
                "name": validation_definition_suite.name,
                "id": validation_definition_suite.id,
            },
            "id": None,
        }

        validation_definition = ValidationDefinition.parse_obj(serialized_config)
        assert validation_definition.name == self.validation_definition_name
        assert validation_definition.data == batch_definition
        assert validation_definition.suite == validation_definition_suite

    @pytest.mark.unit
    @pytest.mark.parametrize(
        "serialized_config, error_substring",
        [
            pytest.param(
                {
                    "name": validation_definition_name,
                    "data": {
                        "asset": {
                            "name": asset_name,
                            "id": None,
                        },
                        "batch_definition": {
                            "name": batch_definition_name,
                            "id": None,
                        },
                    },
                    "suite": {
                        "name": suite_name,
                        "id": None,
                    },
                    "id": None,
                },
                "data did not contain expected identifiers",
                id="bad_data_format[missing_datasource]",
            ),
            pytest.param(
                {
                    "name": validation_definition_name,
                    "data": {},
                    "suite": {
                        "name": suite_name,
                        "id": None,
                    },
                    "id": None,
                },
                "data did not contain expected identifiers",
                id="bad_data_format[empty_field]",
            ),
            pytest.param(
                {
                    "name": validation_definition_name,
                    "data": {
                        "datasource": {
                            "name": ds_name,
                            "id": None,
                        },
                        "asset": {
                            "name": asset_name,
                            "id": None,
                        },
                        "batch_definition": {
                            "name": batch_definition_name,
                            "id": None,
                        },
                    },
                    "suite": {},
                    "id": None,
                },
                "suite did not contain expected identifiers",
                id="bad_suite_format",
            ),
        ],
    )
    def test_validation_definition_deserialization_bad_format(
        self, serialized_config: dict, error_substring: str
    ):
        with pytest.raises(ValueError, match=f"{error_substring}*."):
            ValidationDefinition.parse_obj(serialized_config)

    @pytest.mark.unit
    @pytest.mark.parametrize(
        "serialized_config, error_substring",
        [
            pytest.param(
                {
                    "name": validation_definition_name,
                    "data": {
                        "datasource": {
                            "name": ds_name,
                            "id": None,
                        },
                        "asset": {
                            "name": asset_name,
                            "id": None,
                        },
                        "batch_definition": {
                            "name": batch_definition_name,
                            "id": None,
                        },
                    },
                    "suite": {
                        "name": "i_do_not_exist",
                        "id": None,
                    },
                    "id": None,
                },
                "Could not find suite",
                id="non_existant_suite",
            ),
            pytest.param(
                {
                    "name": validation_definition_name,
                    "data": {
                        "datasource": {
                            "name": "i_do_not_exist",
                            "id": None,
                        },
                        "asset": {
                            "name": asset_name,
                            "id": None,
                        },
                        "batch_definition": {
                            "name": batch_definition_name,
                            "id": None,
                        },
                    },
                    "suite": {
                        "name": suite_name,
                        "id": None,
                    },
                    "id": None,
                },
                "Could not find datasource",
                id="non_existant_datasource",
            ),
            pytest.param(
                {
                    "name": validation_definition_name,
                    "data": {
                        "datasource": {
                            "name": ds_name,
                            "id": None,
                        },
                        "asset": {
                            "name": "i_do_not_exist",
                            "id": None,
                        },
                        "batch_definition": {
                            "name": batch_definition_name,
                            "id": None,
                        },
                    },
                    "suite": {
                        "name": suite_name,
                        "id": None,
                    },
                    "id": None,
                },
                "Could not find asset",
                id="non_existant_asset",
            ),
            pytest.param(
                {
                    "name": validation_definition_name,
                    "data": {
                        "datasource": {
                            "name": ds_name,
                            "id": None,
                        },
                        "asset": {
                            "name": asset_name,
                            "id": None,
                        },
                        "batch_definition": {
                            "name": "i_do_not_exist",
                            "id": None,
                        },
                    },
                    "suite": {
                        "name": suite_name,
                        "id": None,
                    },
                    "id": None,
                },
                "Could not find batch definition",
                id="non_existant_batch_definition",
            ),
        ],
    )
    def test_validation_definition_deserialization_non_existant_resource(
        self,
        validation_definition_data: tuple[PandasDatasource, CSVAsset, BatchDefinition],
        validation_definition_suite: ExpectationSuite,
        serialized_config: dict,
        error_substring: str,
    ):
        with pytest.raises(ValueError, match=f"{error_substring}*."):
            ValidationDefinition.parse_obj(serialized_config)


@pytest.mark.unit
def test_identifier_bundle_no_id_raises_error(validation_definition: ValidationDefinition):
    validation_definition.id = None

    with pytest.raises(ValidationDefinitionRelatedResourcesFreshnessError):
        validation_definition.identifier_bundle()


@pytest.mark.unit
def test_save_success(mocker: MockerFixture, validation_definition: ValidationDefinition):
    context = mocker.Mock(spec=AbstractDataContext)
    set_context(project=context)

    store_key = context.validation_definition_store.get_key.return_value
    validation_definition.save()

    context.validation_definition_store.update.assert_called_once_with(
        key=store_key, value=validation_definition
    )


@pytest.mark.parametrize(
    "has_id,has_suite_id,has_batch_def_id,error_list",
    [
        pytest.param(
            True,
            True,
            True,
            [],
            id="validation_id|suite_id|batch_def_id",
        ),
        pytest.param(
            True,
            False,
            True,
            [ExpectationSuiteNotAddedError],
            id="validation_id|no_suite_id|batch_def_id",
        ),
        pytest.param(
            True,
            True,
            False,
            [BatchDefinitionNotAddedError],
            id="validation_id|suite_id|no_batch_def_id",
        ),
        pytest.param(
            True,
            False,
            False,
            [BatchDefinitionNotAddedError, ExpectationSuiteNotAddedError],
            id="validation_id|no_suite_id|no_batch_def_id",
        ),
        pytest.param(
            False,
            True,
            True,
            [ValidationDefinitionNotAddedError],
            id="no_validation_id|suite_id|batch_def_id",
        ),
        pytest.param(
            False,
            False,
            True,
            [ExpectationSuiteNotAddedError, ValidationDefinitionNotAddedError],
            id="no_validation_id|no_suite_id|batch_def_id",
        ),
        pytest.param(
            False,
            True,
            False,
            [BatchDefinitionNotAddedError, ValidationDefinitionNotAddedError],
            id="no_validation_id|suite_id|no_batch_def_id",
        ),
        pytest.param(
            False,
            False,
            False,
            [
                BatchDefinitionNotAddedError,
                ExpectationSuiteNotAddedError,
                ValidationDefinitionNotAddedError,
            ],
            id="no_validation_id|no_suite_id|no_batch_def_id",
        ),
    ],
)
@pytest.mark.unit
def test_is_fresh(
    in_memory_runtime_context,
    has_id: bool,
    has_suite_id: bool,
    has_batch_def_id: bool,
    error_list: list[Type[ResourceFreshnessError]],
):
    context = in_memory_runtime_context

    batch_definition = (
        context.data_sources.add_pandas(name="my_pandas_ds")
        .add_csv_asset(name="my_csv_asset", filepath_or_buffer="data.csv")
        .add_batch_definition(name="my_batch_def")
    )
    suite = context.suites.add(ExpectationSuite(name="my_suite"))
    validation_definition = context.validation_definitions.add(
        ValidationDefinition(
            name="my_validation_definition",
            suite=suite,
            data=batch_definition,
        )
    )

    # Stores/Fluent API will always assign IDs but we manually override them here
    # for purposes of changing object state for the test
    if not has_batch_def_id:
        validation_definition.data.id = None
    if not has_suite_id:
        validation_definition.suite.id = None
    if not has_id:
        validation_definition.id = None

    diagnostics = validation_definition.is_fresh()
    try:
        diagnostics.raise_for_error()
    except ResourceFreshnessAggregateError as e:
        assert [type(err) for err in e.errors] == error_list


@pytest.mark.unit
def test_is_fresh_raises_error_when_validation_definition_not_found(in_memory_runtime_context):
    context = in_memory_runtime_context

    batch_definition = (
        context.data_sources.add_pandas(name="my_pandas_ds")
        .add_csv_asset(name="my_csv_asset", filepath_or_buffer="data.csv")
        .add_batch_definition(name="my_batch_def")
    )
    suite = context.suites.add(ExpectationSuite(name="my_suite"))
    validation_definition = context.validation_definitions.add(
        ValidationDefinition(
            name="my_validation_definition",
            suite=suite,
            data=batch_definition,
        )
    )

    context.validation_definitions.delete(validation_definition.name)

    diagnostics = validation_definition.is_fresh()
    assert diagnostics.success is False
    assert len(diagnostics.errors) == 1
    assert isinstance(diagnostics.errors[0], ValidationDefinitionNotFoundError)


@pytest.mark.unit
def test_is_fresh_raises_error_when_child_deps_not_found(in_memory_runtime_context):
    context = in_memory_runtime_context

    datasource = context.data_sources.add_pandas(name="my_pandas_ds")
    asset = datasource.add_csv_asset(name="my_csv_asset", filepath_or_buffer="data.csv")
    batch_definition = asset.add_batch_definition(name="my_batch_def")
    suite = context.suites.add(ExpectationSuite(name="my_suite"))
    validation_definition = context.validation_definitions.add(
        ValidationDefinition(
            name="my_validation_definition",
            suite=suite,
            data=batch_definition,
        )
    )

    asset.delete_batch_definition(batch_definition.name)
    context.suites.delete(suite.name)

    diagnostics = validation_definition.is_fresh()
    assert diagnostics.success is False
    assert len(diagnostics.errors) == 2
    assert isinstance(diagnostics.errors[0], BatchDefinitionNotFoundError)
    assert isinstance(diagnostics.errors[1], ExpectationSuiteNotFoundError)



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Literal, Optional
from uuid import UUID

import posthog

from great_expectations.analytics.config import (
    ENV_CONFIG,
    Config,
    update_config,
)

if TYPE_CHECKING:
    from great_expectations.analytics.base_event import Event


def submit(event: Event) -> None:
    """Sends the analytics event to our analytics platform.

    Args:
        event: An object containing the details of the event to be submitted.
    """

    try:
        groups = {
            "data_context": event.data_context_id,
        }

        if event.organization_id:
            groups.update({"organization": event.organization_id})

        posthog.capture(
            str(event.distinct_id),
            str(event.action),
            event.properties(),
            groups=groups,
        )
    except Exception as _:
        # failure to send an analytics event should not be propagated to user
        # TODO: figure out what to do about undeliverable events
        pass


def init(  # noqa: PLR0913 # FIXME CoP
    enable: bool,
    mode: Literal["cloud", "ephemeral", "file"],
    user_id: Optional[UUID] = None,
    data_context_id: Optional[UUID] = None,
    organization_id: Optional[UUID] = None,
    oss_id: Optional[UUID] = None,
    cloud_mode: bool = False,
    user_agent_str: Optional[str] = None,
    remove_profile: bool = True,
):
    """Initializes the analytics platform client."""
    conf = {}
    if user_id:
        conf["user_id"] = user_id
    if data_context_id:
        conf["data_context_id"] = data_context_id
    if organization_id:
        conf["organization_id"] = organization_id
    if oss_id:
        conf["oss_id"] = oss_id
    update_config(
        config=Config(
            cloud_mode=cloud_mode,
            user_agent_str=user_agent_str,
            mode=mode,
            remove_profile=remove_profile,
            **conf,
        )
    )

    enable = enable and not _in_gx_ci()
    posthog.disabled = not enable
    if enable:
        posthog.debug = ENV_CONFIG.posthog_debug
        posthog.project_api_key = ENV_CONFIG.posthog_project_api_key
        posthog.host = ENV_CONFIG.posthog_host


def _in_gx_ci() -> bool:
    return (
        # GitHub Actions
        os.getenv("GITHUB_REPOSITORY") == "great-expectations/great_expectations"
        # Azure Pipelines
        or os.getenv("System.TeamProject") == "great_expectations"
    )



"""Utilities to handle optional imports and related warnings e.g. sqlalchemy.

Great Expectations contains support for datasources and data stores that are
not included in the core package by default. Support requires install of
additional packages. To ensure these code paths are not executed when supporting
libraries are not installed, we check for existence of the associated library.

We also consolidate logic for warning based on version number in this module.
"""

from __future__ import annotations

from typing import Any, Literal, NoReturn

from packaging.version import Version

from great_expectations.compatibility.typing_extensions import override


class NotImported:
    def __init__(self, message: str) -> None:
        self.__dict__["gx_error_message"] = message

    def __getattr__(self, attr: str) -> NoReturn:
        raise ModuleNotFoundError(self.__dict__["gx_error_message"])

    @override
    def __setattr__(self, key: str, value: Any) -> NoReturn:
        raise ModuleNotFoundError(self.__dict__["gx_error_message"])

    def __call__(self, *args, **kwargs) -> NoReturn:
        raise ModuleNotFoundError(self.__dict__["gx_error_message"])

    @override
    def __str__(self) -> str:
        return self.__dict__["gx_error_message"]

    def __bool__(self) -> Literal[False]:
        return False


def is_version_greater_or_equal(version: str | Version, compare_version: str | Version) -> bool:
    """Check if the version is greater or equal to the compare_version.

    Args:
        version: Current version.
        compare_version: Version to compare to.

    Returns:
        Boolean indicating if the version is greater or equal to the compare version.
    """
    if isinstance(version, str):
        version = Version(version)
    if isinstance(compare_version, str):
        compare_version = Version(compare_version)

    return version >= compare_version


def is_version_less_than(version: str | Version, compare_version: str | Version) -> bool:
    """Check if the version is less than the compare_version.

    Args:
        version: Current version.
        compare_version: Version to compare to.

    Returns:
        Boolean indicating if the version is less than the compare version.
    """
    if isinstance(version, str):
        version = Version(version)
    if isinstance(compare_version, str):
        compare_version = Version(compare_version)

    return version < compare_version



from tests.integration.test_utils.data_source_config import (
    BigQueryDatasourceTestConfig,
    DatabricksDatasourceTestConfig,
    DataSourceTestConfig,
    PandasDataFrameDatasourceTestConfig,
    PandasFilesystemCsvDatasourceTestConfig,
    PostgreSQLDatasourceTestConfig,
    SnowflakeDatasourceTestConfig,
    SparkFilesystemCsvDatasourceTestConfig,
    SqliteDatasourceTestConfig,
)

PANDAS_DATA_SOURCES: list[DataSourceTestConfig] = [
    PandasFilesystemCsvDatasourceTestConfig(),
    PandasDataFrameDatasourceTestConfig(),
]

SPARK_DATA_SOURCES: list[DataSourceTestConfig] = [
    SparkFilesystemCsvDatasourceTestConfig(),
]

SQL_DATA_SOURCES: list[DataSourceTestConfig] = [
    BigQueryDatasourceTestConfig(),
    DatabricksDatasourceTestConfig(),
    PostgreSQLDatasourceTestConfig(),
    SnowflakeDatasourceTestConfig(),
    SqliteDatasourceTestConfig(),
]



from collections.abc import Sequence

import pandas as pd
import pytest
from sqlalchemy import types as sqlatypes

import great_expectations.expectations as gxe
from great_expectations.core.result_format import ResultFormat
from great_expectations.datasource.fluent.interfaces import Batch
from great_expectations.render import (
    AtomicDiagnosticRendererType,
    RenderedAtomicContent,
    RenderedAtomicValue,
)
from great_expectations.render.renderer_configuration import RendererSchema, RendererValueType
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.test_utils.data_source_config import RedshiftDatasourceTestConfig
from tests.integration.test_utils.data_source_config.base import DataSourceTestConfig
from tests.integration.test_utils.data_source_config.big_query import BigQueryDatasourceTestConfig
from tests.integration.test_utils.data_source_config.databricks import (
    DatabricksDatasourceTestConfig,
)
from tests.integration.test_utils.data_source_config.mssql import MSSQLDatasourceTestConfig
from tests.integration.test_utils.data_source_config.mysql import MySQLDatasourceTestConfig
from tests.integration.test_utils.data_source_config.pandas_data_frame import (
    PandasDataFrameDatasourceTestConfig,
)
from tests.integration.test_utils.data_source_config.pandas_filesystem_csv import (
    PandasFilesystemCsvDatasourceTestConfig,
)
from tests.integration.test_utils.data_source_config.postgres import PostgreSQLDatasourceTestConfig
from tests.integration.test_utils.data_source_config.snowflake import SnowflakeDatasourceTestConfig
from tests.integration.test_utils.data_source_config.spark_filesystem_csv import (
    SparkFilesystemCsvDatasourceTestConfig,
)
from tests.integration.test_utils.data_source_config.sqlite import SqliteDatasourceTestConfig

ALL_NONNULL_COL = "all_nonnull"
HALF_NONNULL_COL = "half_nonnull"
MOSTLY_NONNULL_COL = "mostly_nonnull"
NO_NONNULL_COL = "no_nonnull"
STRING_COL = "my_strings"
TEST_COL = "test_col"


# Create a sample dataframe with different proportions of nonnull values
DATA = pd.DataFrame(
    {
        ALL_NONNULL_COL: [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        HALF_NONNULL_COL: [0, 1, 2, 3, 4, None, None, None, None, None],
        MOSTLY_NONNULL_COL: [0, 1, 2, 3, 4, 5, 6, 7, None, None],
        NO_NONNULL_COL: [None, None, None, None, None, None, None, None, None, None],
        STRING_COL: ["foo", "bar", "baz", None, None, "qux", "quux", None, "corge", None],
    },
    dtype="object",
)

# Create an empty dataframe
EMPTY_DATA = pd.DataFrame({TEST_COL: []})

COLUMN_TYPES = {NO_NONNULL_COL: sqlatypes.INTEGER}
try:
    from great_expectations.compatibility.pyspark import types as PYSPARK_TYPES

    SPARK_COLUMN_TYPES = {
        ALL_NONNULL_COL: PYSPARK_TYPES.IntegerType,
        HALF_NONNULL_COL: PYSPARK_TYPES.IntegerType,
        MOSTLY_NONNULL_COL: PYSPARK_TYPES.IntegerType,
        NO_NONNULL_COL: PYSPARK_TYPES.IntegerType,
        STRING_COL: PYSPARK_TYPES.StringType,
    }
except ModuleNotFoundError:
    SPARK_COLUMN_TYPES = {}

ALL_DATA_SOURCES: Sequence[DataSourceTestConfig] = [
    BigQueryDatasourceTestConfig(column_types=COLUMN_TYPES),
    DatabricksDatasourceTestConfig(column_types=COLUMN_TYPES),
    MSSQLDatasourceTestConfig(column_types=COLUMN_TYPES),
    MySQLDatasourceTestConfig(column_types=COLUMN_TYPES),
    PandasDataFrameDatasourceTestConfig(),
    PandasFilesystemCsvDatasourceTestConfig(),
    PostgreSQLDatasourceTestConfig(column_types=COLUMN_TYPES),
    RedshiftDatasourceTestConfig(column_types=COLUMN_TYPES),
    SnowflakeDatasourceTestConfig(column_types=COLUMN_TYPES),
    SparkFilesystemCsvDatasourceTestConfig(column_types=SPARK_COLUMN_TYPES),
    SqliteDatasourceTestConfig(column_types=COLUMN_TYPES),
]

# Define a data source config that only includes SQLite
JUST_SQLITE_DATA_SOURCES: Sequence[DataSourceTestConfig] = [
    SqliteDatasourceTestConfig(column_types=COLUMN_TYPES),
]


@parameterize_batch_for_data_sources(data_source_configs=ALL_DATA_SOURCES, data=DATA)
def test_success_complete_results(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnProportionOfNonNullValuesToBeBetween(
        column=HALF_NONNULL_COL, min_value=0.4, max_value=0.7
    )
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    assert result.success
    assert result.to_json_dict()["result"] == {"observed_value": 0.5}


@parameterize_batch_for_data_sources(data_source_configs=ALL_DATA_SOURCES, data=DATA)
def test_strings(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnProportionOfNonNullValuesToBeBetween(
        column=STRING_COL, min_value=0.4, max_value=0.7
    )
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnProportionOfNonNullValuesToBeBetween(column=HALF_NONNULL_COL),
            id="vacuously_true",
        ),
        pytest.param(
            gxe.ExpectColumnProportionOfNonNullValuesToBeBetween(
                column=HALF_NONNULL_COL, min_value=0.4, max_value=0.7
            ),
            id="half_nonnull",
        ),
        pytest.param(
            gxe.ExpectColumnProportionOfNonNullValuesToBeBetween(
                column=ALL_NONNULL_COL, min_value=0.9, max_value=1.0
            ),
            id="all_nonnull",
        ),
        pytest.param(
            gxe.ExpectColumnProportionOfNonNullValuesToBeBetween(
                column=NO_NONNULL_COL, min_value=0.0, max_value=0.0
            ),
            id="no_nonnull",
        ),
        pytest.param(
            gxe.ExpectColumnProportionOfNonNullValuesToBeBetween(
                column=MOSTLY_NONNULL_COL, min_value=0.8
            ),
            id="only_min",
        ),
        pytest.param(
            gxe.ExpectColumnProportionOfNonNullValuesToBeBetween(
                column=MOSTLY_NONNULL_COL, max_value=0.8
            ),
            id="only_max",
        ),
        pytest.param(
            gxe.ExpectColumnProportionOfNonNullValuesToBeBetween(
                column=MOSTLY_NONNULL_COL,
                min_value=0.79,
                max_value=0.81,
                strict_min=True,
                strict_max=True,
            ),
            id="strict_bounds",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_SQLITE_DATA_SOURCES, data=DATA)
def test_success(
    batch_for_datasource: Batch, expectation: gxe.ExpectColumnProportionOfNonNullValuesToBeBetween
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnProportionOfNonNullValuesToBeBetween(
                column=HALF_NONNULL_COL, min_value=0.5, strict_min=True
            ),
            id="strict_min",
        ),
        pytest.param(
            gxe.ExpectColumnProportionOfNonNullValuesToBeBetween(
                column=HALF_NONNULL_COL, max_value=0.5, strict_max=True
            ),
            id="strict_max",
        ),
        pytest.param(
            gxe.ExpectColumnProportionOfNonNullValuesToBeBetween(
                column=HALF_NONNULL_COL, min_value=0.6, max_value=0.7
            ),
            id="wrong_bounds",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_SQLITE_DATA_SOURCES, data=DATA)
def test_failure(
    batch_for_datasource: Batch, expectation: gxe.ExpectColumnProportionOfNonNullValuesToBeBetween
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert not result.success


@parameterize_batch_for_data_sources(data_source_configs=JUST_SQLITE_DATA_SOURCES, data=EMPTY_DATA)
def test_empty_dataframe(batch_for_datasource: Batch) -> None:
    """Test that expectations handle empty dataframes correctly."""
    expectation = gxe.ExpectColumnProportionOfNonNullValuesToBeBetween(
        column=TEST_COL, min_value=0.0, max_value=1.0
    )
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    assert result.success

    expectation_zero = gxe.ExpectColumnProportionOfNonNullValuesToBeBetween(
        column=TEST_COL, min_value=0.0, max_value=0.0
    )
    result_zero = batch_for_datasource.validate(
        expectation_zero, result_format=ResultFormat.COMPLETE
    )

    assert result_zero.success
    assert result_zero.to_json_dict()["result"] == {"observed_value": 0}


@parameterize_batch_for_data_sources(data_source_configs=JUST_SQLITE_DATA_SOURCES, data=DATA)
def test_diagnostic_rendering(batch_for_datasource: Batch) -> None:
    """Test that diagnostic rendering works correctly for the expectation."""
    expectation = gxe.ExpectColumnProportionOfNonNullValuesToBeBetween(
        column=HALF_NONNULL_COL,
        min_value=0.4,
        max_value=0.7,
    )
    result = batch_for_datasource.validate(expectation)
    result.render()

    assert result.rendered_content == [
        RenderedAtomicContent(
            name=AtomicDiagnosticRendererType.OBSERVED_VALUE,
            value_type="StringValueType",
            value=RenderedAtomicValue(
                schema={"type": "com.superconductive.rendered.string"},
                template="$observed_value",
                params={
                    "observed_value": {
                        "schema": RendererSchema(type=RendererValueType.NUMBER),
                        "value": 0.5,
                    },
                },
            ),
        ),
    ]



from datetime import datetime, timezone

import pandas as pd
import pytest

from great_expectations import get_context
from great_expectations.compatibility.snowflake import SNOWFLAKE_TYPES
from great_expectations.compatibility.sqlalchemy import sqltypes
from great_expectations.expectations import (
    ExpectColumnDistinctValuesToContainSet,
    ExpectColumnSumToBeBetween,
    ExpectColumnValuesToBeBetween,
    ExpectColumnValuesToBeOfType,
)
from tests.integration.test_utils.data_source_config import SnowflakeDatasourceTestConfig
from tests.integration.test_utils.data_source_config.snowflake import SnowflakeBatchTestSetup


class TestSnowflakeDataTypes:
    """This set of tests ensures that we can run expectations against every data
    type supported by Snowflake.

    https://docs.snowflake.com/en/sql-reference/intro-summary-data-types
    """

    COLUMN = "col_a"

    @pytest.mark.snowflake
    def test_number(self):
        column_type = (
            SNOWFLAKE_TYPES.NUMBER
        )  # snowflake specific type equivalent to DECIMAL, NUMERIC
        batch_setup = SnowflakeBatchTestSetup(
            config=SnowflakeDatasourceTestConfig(column_types={self.COLUMN: column_type}),
            data=pd.DataFrame({self.COLUMN: [1, 2, 3, 4]}),
            extra_data={},
            context=get_context(mode="ephemeral"),
        )
        with batch_setup.batch_test_context() as batch:
            result = batch.validate(
                expect=ExpectColumnSumToBeBetween(
                    column=self.COLUMN,
                    min_value=9,
                    max_value=11,
                )
            )
        assert result.success

    @pytest.mark.snowflake
    def test_int(self):
        column_type = sqltypes.INT  # equivalent to INTEGER, BIGINT, SMALLINT, TINYINT, BYTEINT
        batch_setup = SnowflakeBatchTestSetup(
            config=SnowflakeDatasourceTestConfig(column_types={self.COLUMN: column_type}),
            data=pd.DataFrame({self.COLUMN: [1, 2, 3, 4]}),
            extra_data={},
            context=get_context(mode="ephemeral"),
        )
        with batch_setup.batch_test_context() as batch:
            result = batch.validate(
                expect=ExpectColumnSumToBeBetween(
                    column=self.COLUMN,
                    min_value=9,
                    max_value=11,
                )
            )
        assert result.success

    @pytest.mark.snowflake
    def test_float(self):
        column_type = sqltypes.FLOAT  # equivalent to FLOAT4, FLOAT8, DOUBLE, DOUBLE PRECISION, REAL
        batch_setup = SnowflakeBatchTestSetup(
            config=SnowflakeDatasourceTestConfig(column_types={self.COLUMN: column_type}),
            data=pd.DataFrame({self.COLUMN: [1, 2, 3, 4]}),
            extra_data={},
            context=get_context(mode="ephemeral"),
        )
        with batch_setup.batch_test_context() as batch:
            result = batch.validate(
                expect=ExpectColumnSumToBeBetween(
                    column=self.COLUMN,
                    min_value=9,
                    max_value=11,
                )
            )
        assert result.success

    @pytest.mark.snowflake
    def test_varchar(self):
        column_type = sqltypes.VARCHAR  # equivalent to STRING, TEXT
        batch_setup = SnowflakeBatchTestSetup(
            config=SnowflakeDatasourceTestConfig(column_types={self.COLUMN: column_type}),
            data=pd.DataFrame({self.COLUMN: ["a", "b", "c", "d"]}),
            extra_data={},
            context=get_context(mode="ephemeral"),
        )
        with batch_setup.batch_test_context() as batch:
            result = batch.validate(
                expect=ExpectColumnDistinctValuesToContainSet(
                    column=self.COLUMN,
                    value_set=[
                        "a",
                        "b",
                    ],
                )
            )
        assert result.success

    @pytest.mark.snowflake
    def test_char(self):
        column_type = sqltypes.CHAR  # length of 1, equivalent to CHARACTER
        batch_setup = SnowflakeBatchTestSetup(
            config=SnowflakeDatasourceTestConfig(column_types={self.COLUMN: column_type}),
            data=pd.DataFrame({self.COLUMN: ["a", "b", "c", "d"]}),
            extra_data={},
            context=get_context(mode="ephemeral"),
        )
        with batch_setup.batch_test_context() as batch:
            result = batch.validate(
                expect=ExpectColumnDistinctValuesToContainSet(
                    column=self.COLUMN,
                    value_set=[
                        "a",
                        "b",
                    ],
                )
            )
        assert result.success

    @pytest.mark.snowflake
    def test_boolean(self):
        column_type = sqltypes.BOOLEAN
        batch_setup = SnowflakeBatchTestSetup(
            config=SnowflakeDatasourceTestConfig(column_types={self.COLUMN: column_type}),
            data=pd.DataFrame({self.COLUMN: [True, False, True, True]}),
            extra_data={},
            context=get_context(mode="ephemeral"),
        )
        with batch_setup.batch_test_context() as batch:
            result = batch.validate(
                expect=ExpectColumnValuesToBeOfType(column=self.COLUMN, type_="BOOLEAN")
            )
        assert result.success

    @pytest.mark.snowflake
    def test_date(self):
        column_type = sqltypes.DATE
        batch_setup = SnowflakeBatchTestSetup(
            config=SnowflakeDatasourceTestConfig(column_types={self.COLUMN: column_type}),
            data=pd.DataFrame(
                {
                    self.COLUMN: [
                        datetime(year=2021, month=1, day=31, tzinfo=timezone.utc).date(),
                        datetime(year=2022, month=1, day=31, tzinfo=timezone.utc).date(),
                        datetime(year=2023, month=1, day=31, tzinfo=timezone.utc).date(),
                    ]
                }
            ),
            extra_data={},
            context=get_context(mode="ephemeral"),
        )
        with batch_setup.batch_test_context() as batch:
            result = batch.validate(
                expect=ExpectColumnValuesToBeBetween(
                    column=self.COLUMN,
                    min_value=datetime(year=2021, month=1, day=1, tzinfo=timezone.utc).date(),
                    max_value=datetime(year=2024, month=1, day=1, tzinfo=timezone.utc).date(),
                )
            )
        assert result.success

    @pytest.mark.snowflake
    def test_datetime(self):
        column_type = sqltypes.DATETIME
        batch_setup = SnowflakeBatchTestSetup(
            config=SnowflakeDatasourceTestConfig(column_types={self.COLUMN: column_type}),
            data=pd.DataFrame(
                {
                    self.COLUMN: [
                        str(datetime(year=2021, month=1, day=31, tzinfo=timezone.utc)),
                        str(datetime(year=2022, month=1, day=31, tzinfo=timezone.utc)),
                        str(datetime(year=2023, month=1, day=31, tzinfo=timezone.utc)),
                    ]
                }
            ),
            extra_data={},
            context=get_context(mode="ephemeral"),
        )
        with batch_setup.batch_test_context() as batch:
            result = batch.validate(
                expect=ExpectColumnValuesToBeBetween(
                    column=self.COLUMN,
                    min_value=datetime(year=2021, month=1, day=1, tzinfo=timezone.utc),
                    max_value=datetime(year=2024, month=1, day=1, tzinfo=timezone.utc),
                )
            )
        assert result.success

    @pytest.mark.snowflake
    def test_timestamp_tz(self):
        column_type = SNOWFLAKE_TYPES.TIMESTAMP_TZ
        batch_setup = SnowflakeBatchTestSetup(
            config=SnowflakeDatasourceTestConfig(column_types={self.COLUMN: column_type}),
            data=pd.DataFrame(
                {
                    self.COLUMN: [
                        str(datetime(year=2021, month=1, day=31, tzinfo=timezone.utc)),
                        str(datetime(year=2022, month=1, day=31, tzinfo=timezone.utc)),
                        str(datetime(year=2023, month=1, day=31, tzinfo=timezone.utc)),
                    ]
                }
            ),
            extra_data={},
            context=get_context(mode="ephemeral"),
        )
        with batch_setup.batch_test_context() as batch:
            result = batch.validate(
                expect=ExpectColumnValuesToBeBetween(
                    column=self.COLUMN,
                    min_value=datetime(year=2021, month=1, day=1, tzinfo=timezone.utc),
                    max_value=datetime(year=2024, month=1, day=1, tzinfo=timezone.utc),
                )
            )
        assert result.success

    @pytest.mark.snowflake
    def test_timestamp_ntz(self):
        column_type = SNOWFLAKE_TYPES.TIMESTAMP_NTZ
        batch_setup = SnowflakeBatchTestSetup(
            config=SnowflakeDatasourceTestConfig(column_types={self.COLUMN: column_type}),
            data=pd.DataFrame(
                {
                    self.COLUMN: [
                        str(datetime(year=2021, month=1, day=31, tzinfo=timezone.utc)),
                        str(datetime(year=2022, month=1, day=31, tzinfo=timezone.utc)),
                        str(datetime(year=2023, month=1, day=31, tzinfo=timezone.utc)),
                    ]
                }
            ),
            extra_data={},
            context=get_context(mode="ephemeral"),
        )
        with batch_setup.batch_test_context() as batch:
            result = batch.validate(
                expect=ExpectColumnValuesToBeBetween(
                    column=self.COLUMN,
                    min_value=datetime(year=2021, month=1, day=1, tzinfo=timezone.utc),
                    max_value=datetime(year=2024, month=1, day=1, tzinfo=timezone.utc),
                )
            )
        assert result.success

    @pytest.mark.snowflake
    @pytest.mark.xfail(
        strict=True,
        reason="time is not an accepted min/max value parameter, and other date types "
        "(datetime, timestamp, etc) fail in the query.",
    )
    def test_time(self):
        column_type = sqltypes.TIME
        batch_setup = SnowflakeBatchTestSetup(
            config=SnowflakeDatasourceTestConfig(column_types={self.COLUMN: column_type}),
            data=pd.DataFrame(
                {
                    self.COLUMN: [
                        datetime(year=2021, month=1, day=31, tzinfo=timezone.utc).time(),
                        datetime(year=2022, month=1, day=31, tzinfo=timezone.utc).time(),
                        datetime(year=2023, month=1, day=31, tzinfo=timezone.utc).time(),
                    ]
                }
            ),
            extra_data={},
            context=get_context(mode="ephemeral"),
        )
        with batch_setup.batch_test_context() as batch:
            result = batch.validate(
                expect=ExpectColumnValuesToBeBetween(
                    column=self.COLUMN,
                    min_value=datetime(year=2021, month=1, day=1, tzinfo=timezone.utc).time(),
                    max_value=datetime(year=2024, month=1, day=1, tzinfo=timezone.utc).time(),
                )
            )
        assert result.success



from datetime import datetime, timezone
from typing import Sequence
from uuid import uuid4

import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.core.result_format import ResultFormat
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.test_utils.data_source_config import (
    BigQueryDatasourceTestConfig,
    DatabricksDatasourceTestConfig,
    DataSourceTestConfig,
    # MSSQLDatasourceTestConfig,
    MySQLDatasourceTestConfig,
    PostgreSQLDatasourceTestConfig,
    RedshiftDatasourceTestConfig,
    SnowflakeDatasourceTestConfig,
    SparkFilesystemCsvDatasourceTestConfig,
    # SqliteDatasourceTestConfig,
)

# pandas not currently supported by this Expecatation
ALL_SUPPORTED_DATA_SOURCES: Sequence[DataSourceTestConfig] = [
    BigQueryDatasourceTestConfig(),
    DatabricksDatasourceTestConfig(),
    # MSSQLDatasourceTestConfig(),  # fix me
    MySQLDatasourceTestConfig(),
    PostgreSQLDatasourceTestConfig(),
    RedshiftDatasourceTestConfig(),
    SnowflakeDatasourceTestConfig(),
    SparkFilesystemCsvDatasourceTestConfig(),
    # SqliteDatasourceTestConfig(),  # fix me
]

# spark and big query not currently supported with extra_data, so we can't test JOIN
# pandas not currently supported by this Expecatation
EXTRA_DATA_SUPPORTED_DATA_SOURCES: Sequence[DataSourceTestConfig] = [
    DatabricksDatasourceTestConfig(),
    # MSSQLDatasourceTestConfig(),  # fix me
    MySQLDatasourceTestConfig(),
    PostgreSQLDatasourceTestConfig(),
    RedshiftDatasourceTestConfig(),
    SnowflakeDatasourceTestConfig(),
    # SqliteDatasourceTestConfig(),  # fix me
]

# pandas and spark not currently supporting partitioners
PARTITIONER_SUPPORTED_DATA_SOURCES: Sequence[DataSourceTestConfig] = [
    BigQueryDatasourceTestConfig(),
    DatabricksDatasourceTestConfig(),
    # MSSQLDatasourceTestConfig(),  # fix me
    MySQLDatasourceTestConfig(),
    PostgreSQLDatasourceTestConfig(),
    RedshiftDatasourceTestConfig(),
    SnowflakeDatasourceTestConfig(),
    # SqliteDatasourceTestConfig(),  # fix me
]

# spark and big query not currently supported with extra_data, so we can't test JOIN
# pandas and spark not currently supporting partitioners
PARTITIONER_AND_EXTRA_DATA_SUPPORTED_DATA_SOURCES: Sequence[DataSourceTestConfig] = [
    DatabricksDatasourceTestConfig(),
    # MSSQLDatasourceTestConfig(),  # fix me
    MySQLDatasourceTestConfig(),
    PostgreSQLDatasourceTestConfig(),
    RedshiftDatasourceTestConfig(),
    SnowflakeDatasourceTestConfig(),
    # SqliteDatasourceTestConfig(),  # fix me
]

TABLE_1 = pd.DataFrame(
    {
        "entity_id": [1, 2],
        "created_at": [
            datetime(year=2024, month=12, day=1, tzinfo=timezone.utc).date(),
            datetime(year=2024, month=11, day=30, tzinfo=timezone.utc).date(),
        ],
        "quantity": [1, 2],
        "temperature": [75, 92],
        "color": ["red", "red"],
    }
)

TABLE_2 = pd.DataFrame(
    {
        "entity_id": [1, 2],
        "created_at": [
            datetime(year=2024, month=12, day=1, tzinfo=timezone.utc).date(),
            datetime(year=2024, month=11, day=30, tzinfo=timezone.utc).date(),
        ],
        "total_quantity": [1, 2],
    }
)

DATE_COLUMN = "created_at"

SUCCESS_QUERIES = [
    "SELECT * FROM {batch} WHERE quantity > 2",
    "SELECT * FROM {batch} WHERE quantity > 2 AND temperature > 91",
    "SELECT * FROM {batch} WHERE quantity > 2 OR temperature > 92",
    "SELECT * FROM {batch} WHERE quantity > 2 ORDER BY quantity DESC",
    "SELECT color FROM {batch} GROUP BY color HAVING SUM(quantity) > 3",
]

JOIN_SUCCESS_QUERIES = [
    """
     SELECT t1.entity_id, t1.quantity, t2.total_quantity
     FROM {batch} t1
     JOIN table_2 t2 USING (entity_id)
     WHERE t1.quantity <> t2.total_quantity
    """,
    """
     SELECT t1.*, t2.record_count FROM
     (SELECT * FROM {batch} AS batch) AS t1
     JOIN
     (SELECT entity_id, SUM(total_quantity) as total_quantity, COUNT(*) as record_count
      FROM table_2 GROUP BY entity_id) AS t2
     ON t1.entity_id = t2.entity_id
     WHERE t1.quantity <> t2.total_quantity
    """,
]

FAILURE_QUERIES = [
    "SELECT * FROM {batch}",
    "SELECT * FROM {batch} WHERE quantity > 0",
    "SELECT * FROM {batch} WHERE quantity > 0 AND temperature > 74",
    "SELECT * FROM {batch} WHERE quantity > 0 OR temperature > 92",
    "SELECT * FROM {batch} WHERE quantity > 0 ORDER BY quantity DESC",
    "SELECT color FROM {batch} GROUP BY color HAVING SUM(quantity) > 0",
]

JOIN_FAILURE_QUERIES = [
    """
     SELECT t1.entity_id, t1.quantity, t2.total_quantity
     FROM {batch} t1
     JOIN table_2 t2 USING (entity_id)
     WHERE t1.quantity = t2.total_quantity
    """,
    """
     SELECT t1.*, t2.record_count FROM
     (SELECT * FROM {batch} AS batch) AS t1
     JOIN
     (SELECT entity_id, SUM(total_quantity) as total_quantity, COUNT(*) as record_count
      FROM table_2 GROUP BY entity_id) AS t2
     ON t1.entity_id = t2.entity_id
     WHERE t1.quantity = t2.total_quantity
    """,
]


@parameterize_batch_for_data_sources(
    data_source_configs=ALL_SUPPORTED_DATA_SOURCES,
    data=TABLE_1,
)
@pytest.mark.parametrize("unexpected_rows_query", SUCCESS_QUERIES)
def test_unexpected_rows_expectation_batch_keyword_success(
    batch_for_datasource,
    unexpected_rows_query,
) -> None:
    expectation = gxe.UnexpectedRowsExpectation(
        description="Expect query with {batch} keyword to succeed",
        unexpected_rows_query=unexpected_rows_query,
    )
    result = batch_for_datasource.validate(expectation)
    assert result.success
    assert result.exception_info.get("raised_exception") is False


@parameterize_batch_for_data_sources(
    data_source_configs=EXTRA_DATA_SUPPORTED_DATA_SOURCES,
    data=TABLE_1,
    extra_data={"table_2": TABLE_2},
)
def test_unexpected_rows_expectation_join_keyword_success(
    batch_for_datasource,
    extra_table_names_for_datasource,
) -> None:
    for join_success_query in JOIN_SUCCESS_QUERIES:
        unexpected_rows_query = join_success_query.replace(
            "table_2", extra_table_names_for_datasource["table_2"]
        )
        expectation = gxe.UnexpectedRowsExpectation(
            description="Expect query with JOIN keyword to succeed",
            unexpected_rows_query=unexpected_rows_query,
        )
        result = batch_for_datasource.validate(expectation)
        assert result.success
        assert result.exception_info.get("raised_exception") is False


@parameterize_batch_for_data_sources(
    data_source_configs=ALL_SUPPORTED_DATA_SOURCES,
    data=TABLE_1,
)
@pytest.mark.parametrize("unexpected_rows_query", FAILURE_QUERIES)
def test_unexpected_rows_expectation_batch_keyword_failure(
    batch_for_datasource,
    unexpected_rows_query,
) -> None:
    expectation = gxe.UnexpectedRowsExpectation(
        description="Expect query with {batch} keyword to fail",
        unexpected_rows_query=unexpected_rows_query,
    )
    result = batch_for_datasource.validate(expectation)
    assert result.success is False
    assert result.exception_info.get("raised_exception") is False


@parameterize_batch_for_data_sources(
    data_source_configs=EXTRA_DATA_SUPPORTED_DATA_SOURCES,
    data=TABLE_1,
    extra_data={"table_2": TABLE_2},
)
def test_unexpected_rows_expectation_join_keyword_failure(
    batch_for_datasource,
    extra_table_names_for_datasource,
) -> None:
    for join_failure_query in JOIN_FAILURE_QUERIES:
        unexpected_rows_query = join_failure_query.replace(
            "table_2", extra_table_names_for_datasource["table_2"]
        )
        expectation = gxe.UnexpectedRowsExpectation(
            description="Expect query with JOIN keyword to fail",
            unexpected_rows_query=unexpected_rows_query,
        )
        result = batch_for_datasource.validate(expectation)
        assert result.success is False
        assert result.exception_info.get("raised_exception") is False


@parameterize_batch_for_data_sources(
    data_source_configs=PARTITIONER_SUPPORTED_DATA_SOURCES,
    data=TABLE_1,
)
@pytest.mark.parametrize("unexpected_rows_query", SUCCESS_QUERIES)
def test_unexpected_rows_expectation_batch_keyword_partitioner_success(
    asset_for_datasource,
    unexpected_rows_query,
) -> None:
    batch = asset_for_datasource.add_batch_definition_monthly(
        name="my-batch-def", column=DATE_COLUMN
    ).get_batch()
    expectation = gxe.UnexpectedRowsExpectation(
        description="Expect query with {batch} keyword and paritioner defined to succeed",
        unexpected_rows_query=unexpected_rows_query,
    )
    result = batch.validate(expectation)
    assert result.success
    assert result.exception_info.get("raised_exception") is False


@parameterize_batch_for_data_sources(
    data_source_configs=PARTITIONER_AND_EXTRA_DATA_SUPPORTED_DATA_SOURCES,
    data=TABLE_1,
    extra_data={"table_2": TABLE_2},
)
def test_unexpected_rows_expectation_join_keyword_partitioner_success(
    asset_for_datasource,
    extra_table_names_for_datasource,
) -> None:
    batch = asset_for_datasource.add_batch_definition_monthly(
        name="my-batch-def", column=DATE_COLUMN
    ).get_batch()
    for join_success_query in JOIN_SUCCESS_QUERIES:
        unexpected_rows_query = join_success_query.replace(
            "table_2", extra_table_names_for_datasource["table_2"]
        )
        expectation = gxe.UnexpectedRowsExpectation(
            description="Expect query with JOIN keyword and paritioner defined to succeed",
            unexpected_rows_query=unexpected_rows_query,
        )
        result = batch.validate(expectation)
        assert result.success
        assert result.exception_info.get("raised_exception") is False


@parameterize_batch_for_data_sources(
    data_source_configs=PARTITIONER_SUPPORTED_DATA_SOURCES,
    data=TABLE_1,
)
@pytest.mark.parametrize("unexpected_rows_query", FAILURE_QUERIES)
def test_unexpected_rows_expectation_batch_keyword_partitioner_failure(
    asset_for_datasource,
    unexpected_rows_query,
) -> None:
    batch = asset_for_datasource.add_batch_definition_monthly(
        name=str(uuid4()), column=DATE_COLUMN
    ).get_batch()
    expectation = gxe.UnexpectedRowsExpectation(
        description="Expect query with {batch} keyword and partitioner defined to fail",
        unexpected_rows_query=unexpected_rows_query,
    )
    result = batch.validate(expectation)
    assert result.success is False
    assert result.exception_info.get("raised_exception") is False


@parameterize_batch_for_data_sources(
    data_source_configs=PARTITIONER_AND_EXTRA_DATA_SUPPORTED_DATA_SOURCES,
    data=TABLE_1,
    extra_data={"table_2": TABLE_2},
)
def test_unexpected_rows_expectation_join_keyword_partitioner_failure(
    asset_for_datasource,
    extra_table_names_for_datasource,
) -> None:
    batch = asset_for_datasource.add_batch_definition_monthly(
        name=str(uuid4()), column=DATE_COLUMN
    ).get_batch()
    for join_failure_query in JOIN_FAILURE_QUERIES:
        unexpected_rows_query = join_failure_query.replace(
            "table_2", extra_table_names_for_datasource["table_2"]
        )
        expectation = gxe.UnexpectedRowsExpectation(
            description="Expect query with JOIN keyword and paritioner defined to fail",
            unexpected_rows_query=unexpected_rows_query,
        )
        result = batch.validate(expectation)
        assert result.success is False
        assert result.exception_info.get("raised_exception") is False


@parameterize_batch_for_data_sources(
    data_source_configs=[PostgreSQLDatasourceTestConfig(), RedshiftDatasourceTestConfig()],
    data=TABLE_1,
)
def test_success_result_format(batch_for_datasource: Batch) -> None:
    result = batch_for_datasource.validate(
        gxe.UnexpectedRowsExpectation(
            unexpected_rows_query="SELECT * FROM {batch} WHERE entity_id = 123"
        )
    )

    assert result.success
    assert result.result == {
        "observed_value": 0,
        "details": {
            "unexpected_rows": [],
        },
    }


@parameterize_batch_for_data_sources(
    data_source_configs=[PostgreSQLDatasourceTestConfig(), RedshiftDatasourceTestConfig()],
    data=TABLE_1,
)
def test_fail_result_format(batch_for_datasource: Batch) -> None:
    result = batch_for_datasource.validate(
        gxe.UnexpectedRowsExpectation(
            unexpected_rows_query="SELECT * FROM {batch} WHERE entity_id = 2"
        )
    )

    assert not result.success
    assert result.result == {
        "observed_value": 1,
        "details": {
            "unexpected_rows": [
                {
                    "entity_id": 2,
                    "created_at": datetime(year=2024, month=11, day=30, tzinfo=timezone.utc).date(),
                    "quantity": 2,
                    "temperature": 92,
                    "color": "red",
                }
            ],
        },
    }


@parameterize_batch_for_data_sources(
    data_source_configs=ALL_SUPPORTED_DATA_SOURCES,
    data=TABLE_1,
)
@pytest.mark.parametrize("unexpected_rows_query", SUCCESS_QUERIES)
def test_success_with_suite_param_other_table_name_(
    batch_for_datasource: Batch, unexpected_rows_query
) -> None:
    suite_param_key = "test_unexpected_rows_expectation"
    expectation = gxe.UnexpectedRowsExpectation(
        description="Expect query with {batch} keyword to succeed",
        unexpected_rows_query={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: unexpected_rows_query}
    )
    assert result.success



from __future__ import annotations

import enum
import logging

logger = logging.getLogger(__name__)


class MetricDomainTypes(enum.Enum):
    """Enum type, whose members signify the data "Domain", on which a metric can be computed.

    A wide variety of "Domain" types can be defined with applicable metrics associated with their respective "Domain"
    types.  The "Domain" types currently in use (`TABLE`, `COLUMN`, `COLUMN_PAIR`, and `MULTICOLUMN`) are declared here.
    """  # noqa: E501 # FIXME CoP

    TABLE = "table"
    COLUMN = "column"
    COLUMN_PAIR = "column_pair"
    MULTICOLUMN = "multicolumn"



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Optional

from great_expectations.compatibility.typing_extensions import override
from great_expectations.execution_engine import (
    ExecutionEngine,
    PandasExecutionEngine,
    SparkDFExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.metrics.column_aggregate_metric_provider import (
    ColumnAggregateMetricProvider,
    column_aggregate_value,
)
from great_expectations.expectations.metrics.metric_provider import metric_value
from great_expectations.validator.metric_configuration import MetricConfiguration

if TYPE_CHECKING:
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )


class ColumnMostCommonValue(ColumnAggregateMetricProvider):
    metric_name = "column.most_common_value"

    @column_aggregate_value(engine=PandasExecutionEngine)
    def _pandas(cls, column, **kwargs):
        mode_list = list(column.mode().values)
        return mode_list

    @metric_value(engine=SparkDFExecutionEngine)
    def _spark(
        cls,
        execution_engine: SqlAlchemyExecutionEngine,
        metric_domain_kwargs: dict,
        metric_value_kwargs: dict,
        metrics: Dict[str, Any],
        runtime_configuration: dict,
    ):
        column_value_counts = metrics["column.value_counts"]
        return list(column_value_counts[column_value_counts == column_value_counts.max()].index)

    @metric_value(engine=SqlAlchemyExecutionEngine)
    def _sqlalchemy(
        cls,
        execution_engine: SqlAlchemyExecutionEngine,
        metric_domain_kwargs: dict,
        metric_value_kwargs: dict,
        metrics: Dict[str, Any],
        runtime_configuration: dict,
    ):
        column_value_counts = metrics["column.value_counts"]
        return list(column_value_counts[column_value_counts == column_value_counts.max()].index)

    @classmethod
    @override
    def _get_evaluation_dependencies(
        cls,
        metric: MetricConfiguration,
        configuration: Optional[ExpectationConfiguration] = None,
        execution_engine: Optional[ExecutionEngine] = None,
        runtime_configuration: Optional[Dict] = None,
    ):
        """Returns a dictionary of given metric names and their corresponding configuration,
        specifying the metric types and their respective domains"""
        dependencies: dict = super()._get_evaluation_dependencies(
            metric=metric,
            configuration=configuration,
            execution_engine=execution_engine,
            runtime_configuration=runtime_configuration,
        )

        if isinstance(execution_engine, (SparkDFExecutionEngine, SqlAlchemyExecutionEngine)):
            dependencies["column.value_counts"] = MetricConfiguration(
                metric_name="column.value_counts",
                metric_domain_kwargs=metric.metric_domain_kwargs,
                metric_value_kwargs={
                    "sort": "value",
                    "collate": None,
                },
            )

        return dependencies
